#!/usr/bin/env python2.7
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
import ecflow
import argparse # for argument parsing     
import sys

if __name__ == "__main__":
    
    DESC = """Will resume any suspended 'node' who name matches input
              Usage:
                Example1: resume all suspended node whose name matches 'fred' for suite grib_api
                   resume.py --host cca --port 4141 --suite grib_api --name fred
            """    
    PARSER = argparse.ArgumentParser(description=DESC,  
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    PARSER.add_argument('--host', default="localhost",   
                        help="The name of the host machine, defaults to 'localhost'")
    PARSER.add_argument('--port', default="3141",   
                        help="The port on the host, defaults to 3141")
    PARSER.add_argument('--suite',   
                        help="The name of the suite")
    PARSER.add_argument('--name', default="install",   
                        help="The name of the node")
    ARGS = PARSER.parse_args()
    print(ARGS)    
     
    # ===========================================================================
    CL = ecflow.Client(ARGS.host, ARGS.port)
    try:
        CL.ping() 

        # get the incremental changes, and merge with defs stored on the Client 
        CL.sync_local()
        
        # check to see if definition exists in the server
        defs = CL.get_defs()
        if len(defs) == 0 :
            print("No suites found, exiting...")
            sys.exit(0) 
         
        paths_list = []
        node_vec = defs.get_all_nodes()
        for node in node_vec:
            if node.name() != ARGS.name: continue
            if node.get_dstate() != ecflow.DState.suspended: continue

            path = node.get_abs_node_path()
            paths = path.split('/')
            if paths[1] == ARGS.suite:
                paths_list.append(path) 
                
        if len(paths_list) > 0:
            CL.resume(paths_list)

#        # requires ecflow >= 4.0.8
#         suite = defs.find_suite(ARGS.suite)
#         if suite is not None:
#             paths_list = []
#             node_vec = suite.get_all_nodes()
#             for node in node_vec:
#                 if node.name() == ARGS.name and node.get_dstate() == DState.SUSPENDED:
#                     paths_list.append(node.get_abs_node_path())
# 
#             print paths_list
#             if len(paths_list) > 0:
#                  CL.resume(paths_list)
        
    except RuntimeError as ex:
        print("Error: " + str(ex))
        print("Check host and port number are correct.")



#!/usr/bin/env python2
# file deepcode ignore SSHHostKeyVerificationDisabled: <comment the reason here>
from __future__ import print_function
import os
import sys
import time
import getopt
from errno import ENOENT
from stat import S_IFDIR, S_IFREG, S_IFBLK
from sys import argv, exit
try:
  from fuse import FUSE, Operations, LoggingMixIn, FuseOSError, fuse_get_context
except:
  raise Exception("#ERR: fuse is needed\n pip install --user fuse ")

try:
    import ecflow
except:
    loc = "/usr/local/apps/ecflow/current/lib/python2.7/site-packages/ecflow"
    sys.path.append(loc)
    import ecflow

try:
    import cdp
    CDP = True
except:
    CDP = False

README = """
NODE.acr  # STATUS: act que sus sub abo
NODE.att  # attributes

SERVER.log  # mounted log file
SERVER.stats  # info
SERVER.ping
SERVER.zombies

# client:
ARG="--ping" ./.exe
ARG="--stats" ./.exe

"""

sms_status = {
    -1: "unknown",
    0: "unknown",
    1: "suspended",
    2: "complete",
    3: "queued",
    4: "submitted",
    5: "active",
    6: "aborted",
    7: "shutdown",
    8: "halted",
    9: "unknown",
}

sms_flags = (
    'FLAG_FORCE_ABORT',
    'FLAG_USER_EDIT',
    'FLAG_TASK_ABORTED',
    'FLAG_EDIT_FAILED',
    'FLAG_CMD_FAILED',
    'FLAG_NO_SCRIPT',
    'FLAG_KILLED',
    'FLAG_MIGRATED',
    'FLAG_LATE',
    'FLAG_MESSAGE',
    'FLAG_BYRULE',
    'FLAG_QUEUELIMIT',
    'FLAG_WAIT',
    'FLAG_LOCKED',
    'FLAG_ZOMBIE',
    'FLAG_NOT_ENQUEUED',
    'FLAG_MAX',
)

smsrepeat = ('REPEAT_DAY', 'REPEAT_MONTH', 'REPEAT_YEAR',
             'REPEAT_INTEGER', 'REPEAT_STRING', 'REPEAT_FILE', 'REPEAT_DATE',
             'REPEAT_ENUMERATED', 'REPEAT_MAX')

sms_late = ('LATE_SUBMITTED', 'LATE_ACTIVE', 'LATE_COMPLETE',
            'LATE_MAX',)

sms_zombies = ('ZOMBIE_SMS', 'ZOMBIE_USER', 'ZOMBIE_NET',)

sms_zomies_t = ('ZOMBIE_GET', 'ZOMBIE_FOB', 'ZOMBIE_DELETE',
                'ZOMBIE_FAIL', 'ZOMBIE_RESCUE',)

"""
another ecFlow python client example

######### environment setup:
virtualenv venv ; source ./venv/bin/activate; pip install fusepy

python $0 $ECF_HOST $ECF_PORT mnt_point

https://github.com/libfuse/libfuse

commands="   autocancel
clock        complete     cron         date         day          defstatus
edit         endfamily    endsuite     endtask      event        extern
family       inlimit      label        late         limit        meter
        repeat       suite        task              time
today        trigger      "

# status    OK, .XXX file
# suspended OK, additional .sus file
# label     OK new_value
# meter     OK value
# event     OK value
# trigger  evaluate NOK
# complete evaluate NOK
# limit value OK + node_paths NOK
# job, output OK check time stamp
# remote output NOK
# time stamp as status update NOK
# server load stat png history zombies OK
# begin

"""

DEBUG = 1
FILESIZE = 10000
UPDATE = "time"
UPDATE = "always"
UPDATE_INTERVAL = 30  # minutes

ECFLOW_CLIENT = "/usr/local/apps/ecflow/current/bin/ecflow_client"
EXE_CMD = "ecflow_client --port %d --host %s ${ARG:=--help};\n"
ignore = ("bdmv", "inf", )

cmds = {  # "log": "--log=get",
    "history": "--log=get",
    "stats": "--stats",
    "png": "--server_load",
    "why": "--group='get; why %s'",
    "log": None,
    "ping": "--ping",
    # "news": "--news",
    # "url": "--url ${ECF_NAME}",
    # + "=%s.%s.png; " % (self.host, self.port),
    # ) + "cat %s.%s.png" % (self.host, self.port),
    "zombies": "--zombie_get"}

ECF_EXTN = "sms"
ECF_EXTN = "ecf"
exts = {key: None for key in cmds}
task_exts = {
    ECF_EXTN: "script",
    "man": "manual",
    "out": "jobout",
    "job": "job",
    "SUB": None, "pout": "output", "pjob": "job", "log": None,
    "why": None, "exe": None, "url": "None", "README": None,
}
exts.update(task_exts)

statuses = {
            -1: "unknown",
             0: "unknown",
             1: "suspended",
             2: "complete",
             3: "queued",
             4: "submitted",
             5: "active",
             6: "aborted",
             7: "shutdown",
             8: "halted",
             9: "unknown",             }

state3 = ("unk", "que", "sub", "act", "com", "abo", "sus", "hal", "shu", "sus")

def attrs(name):
    cur = os.lstat(name)
    return dict((key, getattr(cur, key)) for key in (
        'st_size', 'st_gid', 'st_uid',
        'st_mode', 'st_mtime', 'st_atime', 'st_ctime', ))


def list_dir(node):
    item = node
    res = []
    if CDP:
        if isinstance(item, cdp.sms_node):
            sms_type = {13: "definition",
                        12: "suite",
                        11: "family",
                        10: "task",
                        32: "alias", }
            if sms_type[item.type] == "task":
                res.extend([".%s" % key for key in task_exts])
            name = item.name
            state = ".%s" % sms_status[item.status][:3]
            if state == ".sus":
                state += ".%s" % sms_status[item.savestate][:3]
            res.extend([name + state])
            # item = node.kids
            # while item:
            #     name = item.name
            #     state = ".%03s" % sms_status[item.status]
            #     item = item.next
    else:
        if isinstance(item, ecflow.Task):
            res.extend([".%s" % key for key in task_exts])
        for deco in item.meters:
            res.extend([".meter.%s.%s" % (deco.name(), deco.value())])
        for deco in item.events:
            res.extend([".event.%s.%s" % (deco.name(), deco.value())])
        name = item.name()
        state = ""
        if node.is_suspended():
            state += ".sus"
        state += ".%03s" % item.get_state()

        res.extend([state[:4] + 'sus'])
    return res


def list_att_sms(item):
    import pprint
    attr = dict()
    ones = (
        'act_no',
        'clock',
        'savedstat',
        'action',
        'flags',
        'modified',
        'status',
        'alias',
        'count',
        'gain',
        'late',
        'name',
        'stime',
        'tryno',
        'autocm',
        'date',
        'restore',
        'text',
        'type'
        'wait', 'btime', 'defstatus', 'nid', 'rid', '.user')

    def add(item, key, one=0):
        if not item:
            return
        # if one or key in ones: attr[key] = item
        if key == 'flags':
            out = []
            for i, flag in enumerate(sms_flags):
                print(item, i, 'flag')
                if item & i:
                    out += [flag]
            attr[key] = (item, out)
        elif key == 'limit':
            attr[key] = {'act_no': item.act_no,
                         'max': item.limit,
                         'name': item.name,
                         'base': item.base,
                         'type': item.limittype,
                         'status': item.status,
                         'min': item.min,
                         'nid': item.nid,
                         'tasks': add(item.tasks, 'list'),
                         'unit': item.unit}

        elif key == 'repeat':
            attr[key] = {'act_no': item.act_no,
                         'end': item.end,
                         'mode': item.mode,
                         'name': item.name,
                         'start': item.start,
                         'status': item.status,
                         'step': item.step,
                         'str': add(item.str, list)}

        elif key == 'math':
            operator_s = (
                "or", "and",
                "eq", "ne",
                "lt", "le", "gt", "ge",
                "+", "-",
                "*", "/", "mod",
                "^",
                "not", "~",
                "(", ")")
            out = dict()
            if one != 0:
                out[one] = None
            if item.left:
                out['left'] = add(item.left, 'math'),
            if item.mtype == 19:
                out['name'] = item.name
            else:
                out['op'] = operator_s[item.mtype - 1]
            if item.right:
                out['right'] = add(item.right, 'math'),
            return out

        elif key in ('time', 'cron', 'today'):
            return {'hour': item.hour,
                    'iscron': item.iscron,
                    'minute': item.minute,
                    'nextime': item.nextime,
                    'lasttime': item.lasttime,
                    'relative': item.relative,
                    'repeat': item.repeat,
                    'today': item.today,
                    'weekdays': item.weekdays,
                    'next': add(item.next, key)}

        elif key in ('trigger', 'complete', ):
            attr[key] = {'math': add(item.math, 'math'),
                         'type': item.type,
                         'status': item.status, }
        elif key == 'list':
            out = []
            while item:
                out += [item.name]
                item = item.next
            return out

        elif key in ('inlimit', 'log'):
            attr[key] = add(item, 'list')

        elif key in ('gvar', 'edit', 'label', 'meter', 'event', 'limit'):
            attr[key] = dict()
            while item:
                attr[key][item.name] = item.value
                item = item.next
        else:
            attr[key] = item
    add(item.genvars, 'gvar')
    add(item.variable, 'edit')
    add(item.label, 'label')
    add(item.meter, 'meter')
    add(item.event, 'event')
    add(item.limit, 'limit')
    add(item.inlimit, 'inlimit')
    add(item.log, 'log')
    add(item.act_no, 'act_no')
    add(item.action, 'action')
    add(item.alias, 'alias')
    add(item.autocm, 'autocm')
    add(item.date, 'date')
    add(item.nid, 'nid')
    add(item.gain, 'gain')
    add(item.clock, 'clock')
    add(item.trigger, 'trigger')
    add(item.complete, 'complete')
    add(sms_status[item.defstatus], 'defstatus')
    add(item.late, 'late')
    add(item.flags, 'flags')
    add(item.repeat, 'repeat')
    # add(sms_status[item.status], 'status')
    add(item.wait, 'wait')

    if 0:
        print ("""# item.btime item.label item.meter item.restore item.stime
 item.genvars item.modified item.parent item.rid item.text item.tryno
 item.user_ptr item.complete item.event item.inlimit item.limit
 item.name item.passwd item.savedstat item.this item.type
 item.variable item.count item.flags item.kids item.log item.next
    item.repeat item.status item.time item.user item.wait""")

    pp = pprint.PrettyPrinter(indent=4)
    return pp.pformat(attr) + "\n"


def list_att(node):
    import pprint
    if node is None:
        return "none"
    if CDP:
        if type(node) == cdp.sms_node:
            return list_att_sms(node)

    attr = dict()

    ecf_type = {ecflow.Defs: "defs",
                ecflow.Suite: "suite",
                ecflow.Family: "family",
                ecflow.Task: "task",
                ecflow.Alias: "alias", }

    try:
        attr['kind'] = ecf_type[type(node)]
    except:
        print("#WAR: what?", type(node))
        attr['kind'] = "unknown"

    if isinstance(node, ecflow.Defs):
        attr['gvar'] = []
        attr['edit'] = []
        for item in node.server_variables:
            attr['gvar'] += [(item.name(), item.value())]
        for item in node.user_variables:
            attr['edit'] += [(item.name(), item.value())]
        pp = pprint.PrettyPrinter(indent=4)
        return pp.pformat(attr)

    elif isinstance(node, ecflow.Suite):
        clock = node.get_clock()
        if clock:
            attr['clock'] = "%s" % clock
        attr['begin'] = node.begun()

    attr['status'] = "%s" % node.get_state()
    try:
        attr['update'] = "%s" % node.get_state_change_time()
    except:
        pass

    if node.is_suspended():
        attr['suspended'] = "true"

    if node.has_time_dependencies():
        attr['has_time_dependencies'] = "%s" % node.has_time_dependencies()

    defstatus = node.get_defstatus()
    if defstatus != ecflow.DState.queued:
        attr['defstatus'] = "%s" % defstatus

    autocancel = node.get_autocancel()
    if autocancel:
        attr['autocancel'] = "%s" % autocancel

    repeat = node.get_repeat()
    if not repeat.empty():
        attr['repeat'] = "%s # value:%s" % (repeat,
                                            repeat.value())

    late = node.get_late()
    if late:
        attr['late'] = "%s" % late

    complete_expr = node.get_complete()
    if complete_expr:
        for part_expr in complete_expr.parts:
            trig = "complete "
            if part_expr.and_expr():
                trig = trig + "-a "
            if part_expr.or_expr():
                trig = trig + "-o "
            attr['complete'] = "%s" % trig + " %s" % \
                part_expr.get_expression() + "\n"
    trigger_expr = node.get_trigger()
    if trigger_expr:
        for part_expr in trigger_expr.parts:
            trig = "trigger "
            if part_expr.and_expr():
                trig = trig + "-a "
            if part_expr.or_expr():
                trig = trig + "-o "
            attr['trigger'] = "%s" % trig + " %s" % \
                part_expr.get_expression() + "\n"

    gvar = ecflow.VariableList()
    node.get_generated_variables(gvar)
    attr['gvar'] = []
    for item in gvar:
        attr['gvar'] += [(item.name(), item.value())]
    attr['edit'] = [(item.name(), item.value())
                    for item in node.variables]

    addit(node.meters, attr, 'meter')
    addit(node.events, attr, 'event')
    addit(node.labels, attr, 'label')
    addit(node.limits, attr, 'limit')
    addit(node.inlimits, attr, 'inlimits')
    if node.has_time_dependencies():
        addit(node.time, attr, 'time')
        addit(node.times, attr, 'cron')
        addit(node.todays, attr, 'today')
        addit(node.dates, attr, 'date')
        addit(node.days, attr, "day")
        addit(node.crons, attr, "cron")
    addit(node.zombies, attr, "zombies")

    pp = pprint.PrettyPrinter(indent=4)
    return pp.pformat(attr) + "\n"

olds = dict()
CYCLE = "45r1"
for key in ("o", "law", "mc", "mofc", "tc3", "compo", "ecompo",
            "e_" + CYCLE, "elaw_" + CYCLE, "emc_" + CYCLE, ):
    olds["/%s/" % key] = "/%s.old/" % key


def list_used_variables(node, ARGS=None):
    """ thanks Avi
    https://confluence.ecmwf.int/pages/viewpage.action?pageId=53513079
    """
    if ARGS is None:
        ARGS = object()
        ARGS.var_name = None
        ARGS.not_value = None

    # make a list of all nodes up the parent hierarchy
    parent_hierarchy = []
    parent_hierarchy.append(node)
    parent = node.get_parent()
    while parent:
        parent_hierarchy.append(parent)
        parent = parent.get_parent()
    # now reverse the list
    parent_hierarchy.reverse()
    # now create a map of all used variables going down the hierarchy.
    # Allow variable lower down the hierarchy to override parent variables
    variable_map = {}
    for node in parent_hierarchy:
        for var in node.variables:
            variable_map[var.name()] = var.value()
    # finally print the used variables
    if ARGS.var_name:
        if ARGS.not_value:
            # use exact match for key
            for key in variable_map:
                if ARGS.var_name == key:
                    if ARGS.not_value != variable_map[key]:
                        print("edit " + key + " '" + variable_map[key] + "'")
        else:
            # use substring match for variable name
            for key in variable_map:
                if ARGS.var_name in key:
                    print("edit " + key + " '" + variable_map[key] + "'")
    else:
        for key in variable_map:
            print("edit " + key + " '" + variable_map[key] + "'")


def set_att(fpath, dct, att=('st_size', ), dft=FILESIZE, ext="job"):
    # if DEBUG: print ("set_att", fpath)

    if os.path.exists(fpath):
        print("'FOUND", fpath, "\n", os.lstat(fpath))
        for key in att:
            if "uid" in att or "gid" in att:
                continue
            dct[key] = getattr(os.lstat(fpath), key)
            print(fpath, key, dct[key])

    elif ".old" not in fpath:
        for key in olds:
            if key in fpath:
                return set_att(fpath.replace(key, olds[key]),
                               dct, att, dft, ext)

    dct['st_size'] = FILESIZE
    return dct


def remote_file(node, fname, res):
    if DEBUG:
        print("####### REMOTE")
    if CDP:
        if type(node) == cdp.sms_node:
            return "SMS!!!"

    if node:
        gvar = ecflow.VariableList()
        node.get_generated_variables(gvar)
        for item in gvar:
            # if DEBUG: print(item)
            if item.name() == "ECF_JOBOUT":
                fname = item.value()
                break

    import paramiko
    import base64
    ssh = paramiko.SSHClient()
    HOST = None
    if not HOST:
        return res
    #key = paramiko.RSAKey(data=base64. decodestring(' '))
    #ssh.get_host_keys().add(HOST, 'ssh-rsa', key)
    ssh.set_missing_host_key_policy(
        paramiko.AutoAddPolicy())
    ssh.connect(HOST, username=os.getenv("USER"))  # , password=)
    cmd = "ls %s.running && tail -20 %s.running || " % (
        fname, fname) + "tail -20 %s" % fname
    stdin, stdout, stderr = ssh.exec_command(cmd)
    if DEBUG:
        print("#MSG: live output: ", HOST, cmd)
    for line in stdout:
        #if DEBUG:
        #    print(line,)
        res += "%s" % line
    return res


# def remote_log_file(host, port, kind):
def remote_log_file(node, fname, res):
    if DEBUG:
        print("####### REMOTE")
    if CDP:
        if type(node) == cdp.sms_node:
            return "SMS!!!"

    HOST, PORT = fname.replace(".log", "").split("@")
    keys = {"3141": "/tmp/localhost.ecf.log",
            }
    ssh.set_missing_host_key_policy(
        paramiko.AutoAddPolicy())
    ssh.connect(HOST, username=os.getenv("USER"))  # , password=)
    cmd = "tail -20 %s " % (keys[port])
    stdin, stdout, stderr = ssh.exec_command(cmd)
    if DEBUG:
        print("#MSG: live output: ", HOST, cmd)
    for line in stdout:
        #if DEBUG:
        #    print(line,)
        res += "%s" % line
    return res


def addit(array, cont, name):
    rc = []
    for item in array:
        load = "%s" % item
        if name == "label":
            load += " # value:%s" % item.new_value()
        elif name in ("meter", "event"):
            load += " # value:%s" % item.value()
        elif name in ("limit", ):
            if item.value() > 0:
                load += " # value:%s" % item.value()
                #if DEBUG:
                #    print(item.node_paths)
                for idx, val in enumerate(item.node_paths):
                    print(idx, val)
                    load += " %s " % val  # ???
        rc.append(load)
    if len(rc) > 0:
        cont[name] = rc


def trunc(res, size, offset):
    # if DEBUG: print("offset", offset, size)
    if len(res) > size:
        return "#TRUNCATED\n" + res[offset:offset + size - 30] + "\n"
    return res + "\n"


class FuseEcflow(LoggingMixIn, Operations):
    '''
    A simple Ecflow python client example
    '''

    def __init__(self, host="localhost", port=31415, path='.'):
        self.host = host
        self.port = int(port)
        self.update = int(time.strftime("%H%M"))
        self.root = path
        self.top = None
        self.client()
        #if DEBUG:
        #    print("#MSG: connected to %s " % host + port)

    def client(self):
        self.client = ecflow.Client(self.host, self.port)
        self.client.sync_local()
        self.defs = self.client.get_defs()
        # print (self.defs)

    def get_node(self, path):
        black = ("/ecflow_client", "/autorun", "/BDMV", "/AACS", "BDSVM",
                 "/RCS")
        for key in black:
            if path == key or key in path:
                return None
        if '@' in path:
            return self.client.get_defs()
        elif '/' == path:
            return self.client.get_defs()
        node = self.client.get_defs().find_abs_node(str(path))
        if node is None:
            raise BaseException(path)
        return node

    def chmod(self, path, mode):
        raise FuseOSError(ENOENT)

    def chown(self, path, uid, gid):
        raise FuseOSError(ENOENT)

    def create(self, path, mode):
        raise FuseOSError(ENOENT)

    def destroy(self, path): pass

    def server_cmd(self, ext, run=1):
        if not ext:
            return ""
        elif not ext in cmds:
            print("#WAR: server cmd: what?", ext)
            return ""
        elif ext in ("log", ):
            res = ""
            # res = remote_log_file(node, fname, res)
            # HOST, PORT = fname.replace(".log", "").split("@")
            keys = {3141: "/tmp/localhost.ecf.log",}
            if self.port in keys:
                res = remote_file(None, keys[self.port], self.host)
            return res  # trunc(res, size, offset)
        elif ext == "png" and not run:
            return " " * FILESIZE
        elif cmds[ext] is None:
            return ""
        cmd = ECFLOW_CLIENT + " --port %d --host %s " % (
            self.port, self.host) + cmds[ext]
        #if DEBUG:
        #    print(cmd)

        import commands
        (rc, res) = commands.getstatusoutput(cmd)

        if ext == "png":
            (rc, res) = commands.getstatusoutput("xdg-open %s.%d.png" % (
                self.host, self.port))
        return res

    def getattr(self, path, fh=None):
        uid, gid, pid = fuse_get_context()
        # if DEBUG: print( "getattr", path, fh)

        st = attrs('.')

        if path == '/':
            st['st_mode'] = 0      # (protection bits)
            st['st_ino'] = 0      # (inode number)
            st['st_dev'] = 0      # (device)
            st['st_nlink'] = 0      # (number of hard links)
            st['st_size'] = 0      # (size of file, in bytes)
            st['st_mode'] = (S_IFDIR | 0750)
            st['st_nlink'] = 2
            return st

        st['st_ctime'] = st['st_mtime'] = st['st_atime'] = time.time()

        if '.' in path:
            st['st_mode'] = (S_IFREG | 0444)
            ext = ""
            try:
                splitted = path.split('.')
                path = splitted[0]
                ext = splitted[-1]
                # if "@" in path: node = self.client.get_defs()
                # else: node = self.client.get_defs().find_abs_node(str(path))
                node = self.get_node(path)

            except Exception as e:
                print("#ERR:", e)
                node = None  # return st

            if ext == "att":
                try:
                    st['st_size'] = len(list_att(node))
                except:
                    st['st_size'] = 0

            elif "README" in path or ".meter." in path or ".event" in path:
                st['st_mode'] = (S_IFREG | 0444)
                st['st_size'] = FILESIZE

            elif ext in ("exe", "url", ):
                st['st_mode'] = (S_IFREG | 0700)

            elif "@" in path:
                res = self.server_cmd(ext, 0)
                st['st_size'] = len(res)

            elif (ext in state3 or "sus" in ext) and node:
                tim = node.get_state_change_time()
                st = set_att(path + ".%s" % ext, st,  # dict(),  # attrs(),
                             ('st_size', 'st_atime', 'st_mtime', 'st_ctime'),
                             ext=tim)

            elif ext in ("job", "out", "SUB", "pjob", "pout"):
                st['st_size'] = FILESIZE
                item = None
                if CDP:
                    if type(node) == cdp.sms_node:
                        pass
                else:
                    gvar = ecflow.VariableList()
                    node.get_generated_variables(gvar)
                    for item in gvar:
                        if item.name() == "ECF_JOB":
                            break
                if item:
                    fname = item.value()
                    if ext == "out":
                        fname = fname.replace(".job", '.')

                    elif ".job0" in fname:
                        for key in olds:
                            if key in fname:
                                fname = fname.replace(key, olds[key]).replace(
                                    ".job0", '.job1')

                    elif ext == "pout":
                        for key in olds:
                            if key in fname:
                                fname.replace(key, olds[key])
                                break
                        fname = fname.replace(".job", '.')

                    elif ext == "SUB":
                        if "job0" in fname:
                            fname.replace("job0", "job1")
                            for key in olds:
                                if key in fname:
                                    fname.replace(key, olds[key])
                                    break
                        fname += ".SUB"

                    elif ext == "pjob":
                        for key in olds:
                            if key in fname:
                                fname.replace(key, olds[key])
                                break
                        fname = fname.replace(".job", '.')
                    st = set_att(fname, st, ('st_size', 'st_atime', 'st_mtime',
                                             'st_ctime'),
                                 ext=ext)

        elif 1:
            st['st_mode'] = (S_IFDIR | 0444)
            st['st_size'] = 1

        else:
            raise FuseOSError(ENOENT)
        return st

    def mkdir(self, path, mode):
        raise FuseOSError(ENOENT)

    def refresh(self):
        curr = int(time.strftime("%H%M"))
        if curr - self.update > UPDATE_INTERVAL or UPDATE == "always":
            self.client.sync_local()
            self.update = curr

    def read(self, path, size, offset, fh):
        ext = None  # "nop"
        if len(path) > 2 and '.' in path:
            res = path.split('.')
            path = res[0]
            ext = res[1]

        self.refresh()

        node = None
        res = ''
        # if DEBUG: print('read', path, size, offset, fh, ext)
        if '@' in path and ext != 'att':
            res = self.server_cmd(ext)
            return trunc(res, size, offset)

        if ext in state3 or "sus" in ext:
            # node = self.client.get_defs().find_abs_node(str(path))
            node = self.get_node(path)
            suspended = ''
            if node.is_suspended():
                suspended = ' # node is suspended'

            tim = node.get_state_change_time()
            st = set_att(path + ".%s" % ext, dict(),  # attrs(),
                         ('st_size', 'st_atime', 'st_mtime', 'st_ctime'),
                         ext=tim)
            return 'node %s state is %s%s # last update: %s\n' % (
                node.get_abs_node_path(),
                node.get_state(), suspended, tim)

        elif ext in ignore:
            return '-ignored-'

        elif "README" in path:
            st = attrs('.')
            st['st_mode'] = (S_IFREG | 0444)
            st['st_size'] = FILESIZE
            return README

        elif ext == "att":
            node = self.get_node(path)
            # print(node)
            res = "%s" % list_att(node)
            return trunc(res, size, offset)

        elif ext and not ext in exts:
            print("#WAR: read what?", ext)
            return "-empty-"

        elif ext in ("exe", ):
            print("#exe", path, ext)
            return EXE_CMD % (self.port, self.host)

        elif ext in ("url", ):
            print("#url", path, ext)
            return EXE_CMD % (self.port, self.host)

        elif ext in ("job", "out", "SUB", "pout", "pjob"):
            if ext in ("pout", "pjob"):
                if ".0" in path:
                    path = path.replace(".0", ".1")
            gvar = None
            node = self.get_node(path)
            if CDP and type(node) == cdp.sms_node:
                pass
            else:
                gvar = ecflow.VariableList()
                node.get_generated_variables(gvar)
                fname = path
                for item in gvar:
                    if item.name() == "ECF_JOB":
                        fname = item.value()
                        break

            subfile = None
            if ".job" in fname:
                subfile = fname + ".sub"
            elif ".SUB" in fname:
                subfile = fname.replace(".SUB", ".sub")
            if os.path.exists(subfile):
                with open(fname + ".sub") as f:
                    res = f.read()

            if ext == "out":
                fname = fname.replace(".job", ".")
            if os.path.exists(fname):
                if ext == "out":
                    jobtime = os.path.getmtime(item.value())
                    outtime = os.path.getmtime(fname)
                    if jobtime > outtime:
                        res += "output older than jobfile, %s\n" % item.value()
                        try:
                            res += remote_file(node, fname, res)
                        except Exception as e:
                            print("#ERR:", e)
                        return trunc(res, size, offset)

                f = open(fname)
                f.seek(offset, 0)
                buf = f.read(size)
                f.close()
                return buf

            elif ext in ("xxout", ):
                try:
                    res += remote_file(node, fname, res)
                except Exception as e:
                    print("#ERR:", e)
                return trunc(res, size, offset)

            elif ext in ("log", ):
                # try:
                res = remote_log_file(node, fname, res)
                # except Exception as e:
                # print(e)
                return trunc(res, size, offset)

            if ext in ("out", "man", "ecf", "job", ):
                try:
                    res = "%s" % self.client.get_file(str(path), exts[ext])
                except Exception as e:
                    print("#ERR:", e)
                    res = "%s" % e
            return trunc(res, size, offset)

        elif ext in ("why", ):
            print("##############", ext)
            cmd = ECFLOW_CLIENT + " --port %d --host %s " % (
                self.port, self.host) + cmds[ext] % path
            if CDP and type(node) == cdp.sms_node:
                return "sms!!!"

            import commands
            (rc, res) = commands.getstatusoutput(cmd)
            return trunc(res, size, offset)

        elif ext in (ECF_EXTN, "man"):
            print("##############", ext)
            if CDP and type(node) == cdp.sms_node:
                return "sms!!!"

            res = "%s" % self.client.get_file(str(path), exts[ext])
            return trunc(res, size, offset)

        else:
            node = self.get_node(path)
        res = ""

        if path == '/':
            if self.top:  # SMS
                item = self.top.kids
                while item:
                    res += "%s " % item.name
                    item = item.next
                return res
            for s in self.defs.suites:
                res += "%s " % s.name()  # ECFLOW
            return res

        elif ext == "exe":
            if self.top:
                return "SMS!!!"
            return "ecflow_client --port ${ECF_PORT:=31415}" + \
                " --host ${ECF_HOST:=localhost} ${ARG:=--help}"

        elif node is None:
            return "-empty-node is none"

        elif 1:
            for s in node.nodes:
                res += "%s\n" % s.name()
            return res

        else:
            raise FuseOSError(ENOENT)

        f = open(path)
        f.seek(offset, 0)
        buf = f.read(size)
        f.close()
        return buf

    def readdir(self, path='/', fh=None):
        ext = None
        if "." in path:
            path, ext = path.split(".")
        if ext and not ext in exts:
            if DEBUG:
                print("#readdir: what?", ext)
            return ["/"]
        # if DEBUG: print ("readdir", path, fh, path, ext)

        self.refresh()

        if ext:
            return [".", "..", "ok",
                    path.replace("/", "_"), ext]
        # node = self.client.get_defs().find_abs_node(str(path))
        node = self.get_node(path)

        if node is None:
            print("#WAR: node is None", path)
            return ['.', '..', ]

        elif path == '/':
            nick = '%s@%d' % (self.host, self.port)
            if self.top:
                res = []
                res += [nick + '.%s' % (sms_status[node.status])]
                s = node.kids
                while s:
                    sus = ""
                    state = ".%.3s" % sms_status[s.status]
                    res += ["%s" % s.name, s.name + state + sus]
                    s = s.next
            else:
                res = ['.', '..', nick + "." + 'att',
                       "ecflow_client.exe", "README"]
                res += [nick + ".%s" % key for key in cmds
                        if key != "why"]
                res += [nick + '.%s' % (node.get_server_state())]
                for s in node.suites:
                    sus = ""
                    if s.is_suspended():
                        sus = ".sus"
                    state = ".%.3s" % s.get_state()
                    res += ["%s" % s.name(), s.name() + state + sus]

        elif self.top:
            res = ['.', '..', ".att", ]
            item = node.kids
            while item:
                res += ["%s" % item.name]
                res += list_dir(item)
                item = item.next
        else:
            res = ['.', '..', ".att", ]
            res += ["%s" % n.name() for n in node.nodes]
            res += ["%s.%.3s" % (n.name(), n.get_state()) for n in node.nodes]
            res += list_dir(node)
        return res

    def readlink(self, path):
        raise FuseOSError(ENOENT)

    def rename(self, old, new):
        raise FuseOSError(ENOENT)

    def rmdir(self, path):
        raise FuseOSError(ENOENT)

    def symlink(self, target, source):
        raise FuseOSError(ENOENT)

    def truncate(self, path, length, fh=None):
        raise FuseOSError(ENOENT)

    def unlink(self, path):
        raise FuseOSError(ENOENT)

    def utimens(self, path, times=None):
        raise FuseOSError(ENOENT)

    def write(self, path, data, offset, fh):
        raise FuseOSError(ENOENT)

NOCDP = 0
source = "/tmp"
try:
    sys.path.append(source)
    import cdp
except:
    NOCDP = 1
    print("###WAR: NO_COP")


class FuseSMS(FuseEcflow):

    def __init__(self, host="localhost", port=31415, path='.'):
        super(FuseSMS, self).__init__(host, port, path)

    def client(self):
        if NOCDP:
            print("NOCDP!")
            return
        rc = cdp.new_int()
        c = cdp.sms_node()
        tree = cdp.sms_info_pointed(c)
        os.putenv("SMS_PROG", "%d" % self.port)
        os.environ['SMS_PROG'] = "%d" % self.port
        cdp.cdp_init0()

        user = os.getenv("USER")
        tout = 60
        passw = "1"
        self.num = -1
        rc = cdp.new_int()
        c = cdp.sms_node()
        s = cdp.sms_info_pointed(c)
        self.handle = None
        if 0:
            cdp.sms_client_login(
                self.host, user, passw, tout, self.num, s, rc)
        if not self.top:
            self.top = cdp.sms_cdp_client_login(
                self.host, user, passw, tout, self.port)

    def get_node(self, path):
        if len(path) > 1 and path[-1] == '/':
            path = path[:-1]
            # print(path)
        # print ("find???", path, self.host, self.port, )
        if "@" in path:
            return cdp.sms_node_find('/', self.top)
        node = cdp.sms_node_find(str(path), self.top)
        if not node:
            print("#ERR", path, self.host, self.port, "NOT found")
        return node

    def refresh(self):
        if NOCDP:
            print("#NOCDP!")
            return
        curr = int(time.strftime("%H%M"))
        if curr - self.update > UPDATE_INTERVAL or UPDATE == "always":
            # rc = cdp.sms_client_news( self.handle, self.num );
            # cdp.sms_cdp_update(self.handle)
            self.client()
            self.update = curr


def usage():
    print('usage: %s <host> <port> <mountpoint>' % argv[0])
    print('usage: %s -h <host> -p <port> -m <mountpoint> -s' % argv[0])

if __name__ == '__main__':
    ECFLOW = True
    PORT = 31415
    HOST = "localhost"
    MOUNT = None
    import sys
    try:
        OPTS, ARGS = getopt.getopt(sys.argv[1:],
                                   "?sh:p:m:n:",
                                   ["help", "sms", "host", "port", "mount",
                                    "node", ])
    except getopt.GetoptError as err:
        print("#ERR: what?", )
        usage()
        sys.exit(2)
    for o, a in OPTS:
        if o in ("-?", "--help"):
            usage()
            sys.exit(2)
        elif o in ("-s", "--sms"):
            ECFLOW = False
            if PORT == 31415:
                PORT == 314159
        elif o in ("-h", "--host", "-n", "--node"):
            HOST = a
        elif o in ("-p", "--port"):
            PORT = a
        elif o in ("-m", "--mount"):
            MOUNT = a

    try:
        a = int(PORT)
    except:
        print("#ERR: port shall be an integer")
        sys.exit(2)

    print(os.system("/usr/bin/fusermount -u %s" % MOUNT))
    if os.system("ping -c 1 %s" % HOST) != 0:
        print("#MSG: host appears not to be up" + HOST)
        sys.exit(2)
    elif not MOUNT:
        usage()
        sys.exit(2)
    else:
        os.system("/usr/bin/fusermount -u %s" % MOUNT)
    if not os.path.exists(MOUNT):
        os.makedirs(MOUNT)
    # elif 1: print os.stat(MOUNT)[os.stat.ST_MODE]
    # elif not os.path.isdir(MOUNT): os.mkdir(MOUNT, 0755)

    if ECFLOW:
        FUS = FUSE(FuseEcflow(HOST, PORT, MOUNT), MOUNT,
                   foreground=True, nothreads=True, )
        # allow_root=True, # echo user_allow_other >> /etc/fuse.conf
        # allow_other=True, )
    else:
        FUS = FUSE(FuseSMS(HOST, PORT, MOUNT), MOUNT,
                   foreground=True, nothreads=True, )

"""stream - class - date - expver - repres - levtype
type - param - step - time - domain - grid - resol

path fullpath queue (suite/j/o) npes threads prio mem

servers id (PK) host port kind
tasks id (PK)
task id (PK) nick FK server FK

event_type job SUB sub act com/abt out event meter id (PK) desc

"""



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# code for testing copy
#
import ecflow
import sys
import os
import copy
import ecflow_test_util as Test

if __name__ == "__main__":
    
    Test.print_test_start(os.path.basename(__file__))

    defs = ecflow.Defs()
    defs.add_suite("a").add_family("f1").add_task("t1").add_variable("a","b").add_event(1).add_meter("meter", 0, 100).add_label("label", "v").add_time("+00:30 20:00 01:00")
    print(defs)
    
    defs_copy = copy.copy(defs)
    print(defs_copy)
    assert defs_copy == defs,"defs should be equal after copy"
    
 
    #===========================================================================
    # Defs: add and delete *USER* variables
    #===========================================================================
    defs = ecflow.Defs()
    defs.add_variable("FRED", "/tmp/")
    defs.add_variable("ECF_URL_CMD", "${BROWSER:=firefox} -new-tab %ECF_URL_BASE%/%ECF_URL%")
    defs.add_variable("ECF_URL_BASE", "http://www.ecmwf.int")
    defs.add_variable("ECF_URL", "publications/manuals/sms")
    assert len(list(defs.user_variables)) == 4, "Expected *user* 4 variable"    
    defs_copy = copy.copy(defs)
    assert len(list(defs_copy.user_variables)) == 4, "Expected *user* 4 variable"    

    defs.delete_variable("FRED");     assert len(list(defs.user_variables)) == 3, "Expected 3 variables since we just delete FRED"
    defs.delete_variable("");         assert len(list(defs.user_variables)) == 0, "Expected 0 variables since we should have deleted all"
    assert len(list(defs_copy.user_variables)) == 4, "Expected *user* 4 variable"    

    #===========================================================================
    # Suite: add and delete variables
    #===========================================================================
    suite = ecflow.Suite("s1")
    suite.add_variable(ecflow.Variable("ECF_HOME", "/tmp/"))
    suite.add_variable("ECF_URL_CMD", "${BROWSER:=firefox} -new-tab %ECF_URL_BASE%/%ECF_URL%")
    suite.add_variable("ECF_URL_BASE", "http://www.ecmwf.int")
    suite.add_variable("ECF_URL", "publications/manuals/sms")
    assert len(list(suite.variables)) == 4, "Expected 4 variable"   
    suite_copy = copy.copy(suite)
    assert len(list(suite_copy.variables)) == 4, "Expected 4 variables in copy"   

    #===========================================================================
    # add and delete limits
    #===========================================================================
    suite.add_limit(ecflow.Limit("limitName1", 10))
    suite.add_limit(ecflow.Limit("limitName2", 10))
    suite.add_limit("limitName3", 10)
    suite.add_limit("limitName4", 10)
    assert len(list(suite.limits)) == 4, "Expected 4 Limits"
    
    suite_copy = copy.copy(suite)
    assert len(list(suite_copy.limits)) == 4, "Expected 4 Limits in copy"
    suite_copy.delete_limit("limitName1"); assert len(list(suite_copy.limits)) == 3, "Expected 3 limits since we just deleted one limitName1" 
    suite_copy.delete_limit("");           assert len(list(suite_copy.limits)) == 0, "Expected 0 limits since we just deleted all of them"

    #===========================================================================
    # Test Limit and node paths 
    #===========================================================================
    the_limit = ecflow.Limit("limitName1", 10)
    assert the_limit.name() == "limitName1", "name not as expected"
    assert the_limit.value() == 0 ,"Expected limit value of 0"
    assert the_limit.limit() == 10 ,"Expected limit of 10"
    assert len(list(the_limit.node_paths())) == 0 ,"Expected nodes which have consumed a limit to be empty"
    
    the_limit.increment(1,"/path1") ;  assert the_limit.value() == 1 ,"Expected limit value of 1"
    the_limit.increment(1,"/path2") ;  assert the_limit.value() == 2 ,"Expected limit value of 2"
    the_limit.increment(1,"/path3") ;  assert the_limit.value() == 3 ,"Expected limit value of 3"
    the_limit.increment(1,"/path4") ;  assert the_limit.value() == 4 ,"Expected limit value of 4"
    for path in the_limit.node_paths(): print(path)
    assert len(list(the_limit.node_paths())) == 4 ,"expected 4 path"
    
    limit_copy = copy.copy(the_limit)
    assert len(list(limit_copy.node_paths())) == 4 ,"expected 4 path in copy"

    limit_copy.decrement(1,"/path1") ; assert limit_copy.value() == 3 ,"Expected limit value of 3"
    limit_copy.decrement(1,"/path2") ; assert limit_copy.value() == 2 ,"Expected limit value of 2"
    limit_copy.decrement(1,"/path3") ; assert limit_copy.value() == 1 ,"Expected limit value of 1"
    limit_copy.decrement(1,"/path4") ; assert limit_copy.value() == 0 ,"Expected limit value of 0"
    assert len(list(limit_copy.node_paths())) == 0 ,"Expected nodes which have consumed a limit to be empty"

    limit_copy.increment(1,"/path1") # add same path, should only consume one token
    limit_copy.increment(1,"/path1")
    limit_copy.increment(1,"/path1")
    limit_copy.increment(1,"/path1")
    assert len(list(limit_copy.node_paths())) == 1 ,"expected 1 path"
    assert limit_copy.value() == 1 ,"Expected limit value of 1"

    #===========================================================================
    # add and delete inlimits
    #===========================================================================
    suite.add_inlimit(ecflow.InLimit("limitName1", "/s1/f1", 2))
    suite.add_inlimit(ecflow.InLimit("limitName2", "/s1/f1", 2, True))
    suite.add_inlimit("limitName3", "/s1/f1", 2)
    suite.add_inlimit("limitName4", "/s1/f1", 2, True)
    assert len(list(suite.inlimits)) == 4, "Expected 4 inLimits"
    
    suite_copy = copy.copy(suite)
    assert len(list(suite_copy.inlimits)) == 4, "Expected 4 inLimits"
    suite_copy.delete_inlimit("limitName1"); assert len(list(suite_copy.inlimits)) == 3, "Expected 3 inlimits since we just deleted one limitName1" 
    suite_copy.delete_inlimit("");           assert len(list(suite_copy.inlimits)) == 0, "Expected 0 inlimits since we just deleted all of them"

    #===============================================================================
    # add and delete triggers and complete
    #===============================================================================
    task = ecflow.Task("task")
    task.add_trigger("t2 == active")
    task.add_complete("t2 == complete")
    
    task_copy = copy.copy(task)
    assert task_copy.get_complete(), "Expected complete"
    assert task_copy.get_trigger(), "Expected trigger"
    task_copy.delete_trigger();  assert not task_copy.get_trigger(), "Expected no trigger"
    task_copy.delete_complete(); assert not task_copy.get_complete(), "Expected no complete"
    
    #===========================================================================
    # Add triggers using expressions
    #===========================================================================
    expr = ecflow.Expression("t1 == complete")
    task = ecflow.Task("task")
    task.add_trigger(expr)
    
    task_copy = copy.copy(task)
    assert task_copy.get_trigger(), "Expected trigger"

    expr = ecflow.Expression("t1 == complete")
    expr.add(ecflow.PartExpression("t1 == complete", True))
    expr.add(ecflow.PartExpression("t2 == complete", True))
    task = ecflow.Task("task")
    task.add_trigger(expr)

    task_copy = copy.copy(task)
    assert task_copy.get_trigger(), "Expected trigger"

     
    #===========================================================================
    # add,delete,find events
    #===========================================================================
    task.add_event(ecflow.Event(1)); 
    task.add_event(2)                              
    task.add_event(ecflow.Event(10, "Eventname"))  
    task.add_event(10, "Eventname2")               
    task.add_event("fred")                         
    
    # test find
    task_copy = copy.copy(task)
    event = task_copy.find_event("EVENT")
    assert(event.empty()),"Expected to not find event"
    assert(event.name() == ""),"Expected to not to find event, number is maximum int"
    assert(event.value() == 0),"Expected to not to find event"

    event = task_copy.find_event("1"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 1), "Expected to find event 1"
    assert(event.name() == ""), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task_copy.find_event("2"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 2), "Expected to find event 1"
    assert(event.name() == ""), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task_copy.find_event("10"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 10), "Expected to find event 10"
    assert(event.name() == "Eventname"), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task_copy.find_event("fred"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.name() == "fred"), "Expected name to be empty, when name defind an not number, number is max_int"
    assert(event.value() == 0),"Expected to not to find event"

    for e in task_copy.events:     print(str(e)," # value: ",str(e.value()))
    a_dict = {}
    for e in task_copy.events:
        if e.name() != "": a_dict[e.name()] = e.value()
        else:              a_dict[e.number()] = e.value()
    print(a_dict)

    assert len(list(task_copy.events)) == 5, "Expected 5 Events"
    task_copy.delete_event("1");         assert len(list(task_copy.events)) == 4, "Expected 4 Events"
    task_copy.delete_event("Eventname"); assert len(list(task_copy.events)) == 3, "Expected 3 Events"
    task_copy.delete_event("");          assert len(list(task_copy.events)) == 0, "Expected 0 Events"


    #===========================================================================
    # add and delete meter
    #===========================================================================
    task.add_meter(ecflow.Meter("metername1", 0, 100, 50))
    task.add_meter(ecflow.Meter("metername2", 0, 100))
    task.add_meter("metername3", 0, 100, 50)
    task.add_meter("metername4", 0, 100)
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.meters)) == 4, "Expected 4 Meters"
    task_copy.delete_meter("metername1"); assert len(list(task_copy.meters)) == 3, "Expected 3 Meters"
    task_copy.delete_meter("metername4"); assert len(list(task_copy.meters)) == 2, "Expected 2 Meters"
    task_copy.delete_meter("");           assert len(list(task_copy.meters)) == 0, "Expected 0 Meters"

    #===========================================================================
    # add and delete queue
    #===========================================================================
    queue_items = [ "a", "b"]
    task.add_queue(ecflow.Queue("q1", queue_items ))
    task.add_queue(ecflow.Queue("q2", queue_items))
    task.add_queue("q3", [ "a", "b"])
    task.add_queue("q4",  [ "a", "b"])
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.queues)) == 4, "Expected 4 queues"
    task_copy.delete_queue("q1"); assert len(list(task_copy.queues)) == 3, "Expected 3 queues"
    task_copy.delete_queue("q2"); assert len(list(task_copy.queues)) == 2, "Expected 2 queues"
    task_copy.delete_queue("");   assert len(list(task_copy.queues)) == 0, "Expected 0 queues"

    #===========================================================================
    # copy generic
    #===========================================================================
    generic_items = [ "a", "b"]
    task.add_generic(ecflow.Generic("q1",generic_items ))
    task.add_generic(ecflow.Generic("q2",generic_items))
    task.add_generic("q3", [ "a", "b"])
    task.add_generic("q4",  [ "a", "b"])
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.generics)) == 4, "Expected 4  Generics after copy"
    task_copy.delete_generic("q1"); assert len(list(task_copy.generics)) == 3, "Expected 3 Generics"
    task_copy.delete_generic("q2"); assert len(list(task_copy.generics)) == 2, "Expected 2 Generics"
    task_copy.delete_generic("");   assert len(list(task_copy.generics)) == 0, "Expected 0 Generics"

    #===========================================================================
    # add and delete label
    #===========================================================================
    task.add_label(ecflow.Label("label_name1", "value"))
    task.add_label(ecflow.Label("label_name2", "value"))
    task.add_label("label_name3", "value")
    task.add_label("label_name4", "value")
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.labels)) == 4, "Expected 4 labels"
    task_copy.delete_label("label_name1"); assert len(list(task_copy.labels)) == 3, "Expected 3 Labels"
    task_copy.delete_label("label_name4"); assert len(list(task_copy.labels)) == 2, "Expected 2 Labels"
    task_copy.delete_label("");            assert len(list(task_copy.labels)) == 0, "Expected 0 Labels"


    #===========================================================================
    # add delete Repeat
    #===========================================================================
    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatInteger("integer", 0, 100, 2))
    task_copy = copy.copy(task)
    repeat = task_copy.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task_copy.delete_repeat()      
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatEnumerated("enum", ["red", "green", "blue" ]))
    task_copy = copy.copy(task)
    task_copy.delete_repeat()      
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatDate("date", 20100111, 20100115, 2))
    task_copy = copy.copy(task)
    task_copy.delete_repeat()      
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatDateTime("datetime", "20100111T000000", "20100115T000000", "48:00:00"))
    task_copy = copy.copy(task)
    task_copy.delete_repeat()
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatDateList("date", [20100111, 20100115]))
    task_copy = copy.copy(task)
    task_copy.delete_repeat()      
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    task = ecflow.Task("task")
    task.add_repeat(ecflow.RepeatString("string", ["a", "b", "c" ]))
    task_copy = copy.copy(task)
    task_copy.delete_repeat()      
    repeat = task_copy.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    
    #===========================================================================
    # create a time series, used for adding time and today
    #===========================================================================
    start = ecflow.TimeSlot(0, 0)
    finish = ecflow.TimeSlot(23, 0)
    incr = ecflow.TimeSlot(0, 30)
    time_series = ecflow.TimeSeries(start, finish, incr, True)

    # Add and delete today
    # NOTE: **********************************************************************
    # Do *NOT* delete attributes using iterator traversal
    # The iterators *are* bound to the c++ iterators hence if we delete
    # over the traversal, we can corrupt the vector, leading to undefined behavour.  
    # Hence we **cannot** delete more than once over a traversal
    # Soln 1: take a copy
    # *****************************************************************************
    # task.add_today( Today( 0,10 ))
    # task.add_today( Today( 0,59, True ))
    # task.add_today( Today(TimeSlot(20,10)) )
    # task.add_today( Today(TimeSlot(20,20),False)) 
    # for today in task.todays: 
    #    task.delete_today(today) # will corrupt C++ vector
  
    task.add_today("00:30")
    task.add_today("+00:30")
    task.add_today("+00:30 20:00 01:00")
    task.add_today(ecflow.Today(time_series))
    task.add_today(ecflow.Today(0, 10))
    task.add_today(0, 59, True)
    task.add_today(ecflow.Today(ecflow.TimeSlot(20, 10)))
    task.add_today(ecflow.Today(ecflow.TimeSlot(20, 20), False)) 
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.todays)) == 8, "Expected 8 todays"
    vec_copy = []
    for today in task_copy.todays: vec_copy.append(today)
    for today in vec_copy: task_copy.delete_today(today)
    assert len(list(task_copy.todays)) == 0, "Expected 0 todays"
        
    #===========================================================================
    # add and delete time
    #===========================================================================
    task.add_time("00:30")
    task.add_time("+00:30")
    task.add_time("+00:30 20:00 01:00")
    task.add_time(ecflow.Time(time_series))
    task.add_time(ecflow.Time(0, 10))
    task.add_time(0, 59, True)
    task.add_time(ecflow.Time(ecflow.TimeSlot(20, 10)))
    task.add_time(ecflow.Time(ecflow.TimeSlot(20, 20), False)) 

    task_copy = copy.copy(task)
    assert len(list(task_copy.times)) == 8, "Expected 8 times"
    vec_copy = []
    for time in task_copy.times: vec_copy.append(time)
    for time in vec_copy: task_copy.delete_time(time)
    assert len(list(task_copy.times)) == 0, "Expected 0 todays"

    #===========================================================================
    # add and delete date
    #===========================================================================
    for i in [ 1, 2, 4, 8, 16 ] :
        task.add_date(i, 0, 0)
    task.add_date(ecflow.Date(1, 1, 2010))
    task.add_date(1, 1, 2010)       # duplicate
    task.add_date(ecflow.Date(2, 1, 2010))
    task.add_date(ecflow.Date(3, 1, 2010))
    task.add_date(ecflow.Date(4, 1, 2010)) 
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.dates)) == 10, "Expected 10 dates but found " + str(len(list(task_copy.dates)))
    vec_copy = []
    for attr in task_copy.dates: vec_copy.append(attr)
    for attr in vec_copy: task_copy.delete_date(attr)
    assert len(list(task_copy.dates)) == 0, "Expected 0 dates"

    #===========================================================================
    # add and delete day
    #===========================================================================
    task.add_day(ecflow.Day(ecflow.Days.sunday))
    task.add_day(ecflow.Days.monday)
    task.add_day(ecflow.Days.tuesday)  # duplicate ?
    task.add_day("sunday")     
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.days)) == 4, "Expected 4 days"
    vec_copy = []
    for attr in task_copy.days: vec_copy.append(attr)
    for attr in vec_copy: task_copy.delete_day(attr)
    assert len(list(task_copy.days)) == 0, "Expected 0 days"
    
    #===========================================================================
    # add and delete crons
    #===========================================================================
    cron = ecflow.Cron()
    start = ecflow.TimeSlot(23 , 0)
    ts = ecflow.TimeSeries(start,True)  # True means relative to suite start
    cron.set_time_series(ts)
    
    cron = ecflow.Cron()
    cron.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron.set_days_of_month([1, 2, 3, 4, 5, 6 ])
    cron.set_months([1, 2, 3, 4, 5, 6])
    start = ecflow.TimeSlot(0 , 0)
    finish = ecflow.TimeSlot(23, 0)
    incr = ecflow.TimeSlot(0, 30)
    ts = ecflow.TimeSeries(start, finish, incr, True)  # True means relative to suite start
    cron.set_time_series(ts)

    cron0 = ecflow.Cron()
    cron0.set_week_days([0, 1, 2, 3, 4, 5 ])
    cron0.set_time_series(1, 30) # default relative = false, added in release 4.0.7

    cron1 = ecflow.Cron()
    cron1.set_week_days([0, 1, 2, 3, 4, 5, 6 ])
    cron1.set_time_series(1, 30, True)
    
    cron2 = ecflow.Cron()
    cron2.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron2.set_time_series("00:30 01:30 00:01")
    
    cron3 = ecflow.Cron()
    cron3.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron3.set_time_series("+00:30")

    task = ecflow.Task("task")
    task.add_cron(cron)
    task.add_cron(cron1)
    task.add_cron(cron2)
    task.add_cron(cron3)
    
    task_copy = copy.copy(task)
    assert len(list(task_copy.crons)) == 4, "Expected 4 crons"
    vec_copy = []
    for attr in task_copy.crons: vec_copy.append(attr)
    for attr in vec_copy: task_copy.delete_cron(attr)
    assert len(list(task_copy.crons)) == 0, "Expected 0 crons"  
    
    #===========================================================================
    # add autocancel
    #===========================================================================
    print("test add autoCancel")
    t1 = ecflow.Task("t1")
    assert t1.get_autocancel() is None, " Expected no autocancel"
    t1.add_autocancel(3)  # 3 days
    
    task_copy = copy.copy(t1)
    assert task_copy.get_autocancel() is not None, " Expected autocancel"
    print(str(task_copy.get_autocancel()))

    #===========================================================================
    # add late  
    #===========================================================================
    late = ecflow.Late()
    late.submitted(ecflow.TimeSlot(20, 10))
    late.active(ecflow.TimeSlot(20, 10))
    late.complete(ecflow.TimeSlot(20, 10), True)
    task.add_late(late)
    task_copy = copy.copy(task)
    assert task_copy.get_late() is not None, " Expected late"
    
    #===========================================================================
    # add defstatus, last one set takes effect
    #===========================================================================
    task.add_defstatus(ecflow.DState.complete)
    task_copy = copy.copy(task)
    assert task_copy.get_defstatus() == ecflow.DState.complete

    #===========================================================================
    # add clock
    #===========================================================================
    clock = ecflow.Clock(1, 1, 2010, False)     # day,month, year, hybrid
    clock.set_gain(1, 10, True)                 # True means positive gain
    suite = ecflow.Suite("suite")
    assert suite.get_clock() is None, "Expected no clock"
    suite.add_clock(clock)
    
    suite_copy = copy.copy(suite)
    assert suite_copy.get_clock() is not None, "expected clock"
    
    #===========================================================================
    # Add zombie. Note we can *NOT* add two zombie attributes of the same ZombieType
    #===========================================================================
    s1 = ecflow.Task("s1")
    zombie_life_time_in_server = 800
    child_list = [ ecflow.ChildCmdType.init, ecflow.ChildCmdType.event, ecflow.ChildCmdType.meter, ecflow.ChildCmdType.label, ecflow.ChildCmdType.wait, ecflow.ChildCmdType.abort, ecflow.ChildCmdType.complete ]
    zombie_type_list = [ ecflow.ZombieType.ecf, ecflow.ZombieType.ecf_pid, ecflow.ZombieType.ecf_pid_passwd, ecflow.ZombieType.ecf_passwd, ecflow.ZombieType.user, ecflow.ZombieType.path ]
    for zombie_type in zombie_type_list:
        zombie_attr = ecflow.ZombieAttr(zombie_type, child_list, ecflow.ZombieUserActionType.block, zombie_life_time_in_server)
        s1.add_zombie(zombie_attr)
    task_copy = copy.copy(s1)
    assert len(list(task_copy.zombies)) == 6, "Expected 6 zombie attributes but found " + str(len(list(task_copy.zombies)))
    
    # delete all the zombies
    s1.delete_zombie("")
    assert len(list(s1.zombies)) == 0, "Expected zero zombie attributes but found " + str(len(list(s1.zombies)))
    
    # add with zombie_life_time_in_server not set, this is optional
    for zombie_type in zombie_type_list:
        zombie_attr = ecflow.ZombieAttr(zombie_type, child_list, ecflow.ZombieUserActionType.block)
        s1.add_zombie(zombie_attr)
    task_copy = copy.copy(s1)
    assert len(list(task_copy.zombies)) == 6, "Expected 6 zombie attributes but found " + str(len(list(task_copy.zombies)))

    s1.delete_zombie("")
    assert len(list(s1.zombies)) == 0, "Expected zero zombie attributes but found " + str(len(list(s1.zombies)))

    print("All Tests pass")



import os
import ecflow

defs = ecflow.Defs()
suite = defs.add_suite("operation_suite")
suite.add_repeat(ecflow.RepeatDay(1))
suite.add_variable("ECF_HOME", os.getenv("HOME") + "/course")
suite.add_variable("ECF_INCLUDE", os.getenv("HOME") + "/course")
suite.add_variable("ECF_FILES", os.getenv("HOME") + "/course/oper")

# Defines the triggers for the first cycle
cycle_triggers = "1"
for cycle in ("00", "12"):

    if cycle == "12":
        last_step = 240
    else:
        last_step = 24

    fcycle_fam = suite.add_family(cycle)
    fcycle_fam.add_variable("CYCLE", cycle)
    fcycle_fam.add_variable("LAST_STEP", last_step)

    if cycle_triggers != "1":
        fcycle_fam.add_trigger(cycle_triggers)

    analysis_fam = fcycle_fam.add_family("analysis")
    analysis_fam.add_task("get_observations")
    analysis_fam.add_task("run_analysis").add_trigger("get_observations == complete")
    analysis_fam.add_task("post_processing").add_trigger("run_analysis == complete")

    forecast_fam = fcycle_fam.add_family("forecast")
    forecast_fam.add_trigger("analysis == complete")
    forecast_fam.add_task("get_input_data")
    run_forecast_task = forecast_fam.add_task("run_forecast")
    run_forecast_task.add_trigger("get_input_data == complete")
    run_forecast_task.add_meter("step", 0, last_step, last_step)

    archive_fam = fcycle_fam.add_family("archive")
    fam_analsis = archive_fam.add_family("analysis")
    fam_analsis.add_variable("TYPE", "analysis")
    fam_analsis.add_variable("STEP", "0")
    fam_analsis.add_trigger("../analysis/run_analysis == complete")
    fam_analsis.add_task("save")

    for i in range(6, last_step + 1, 6):
        step_fam = fam_analsis.add_family("step_" + str(i))
        step_fam.add_variable("TYPE", "forecast")
        step_fam.add_variable("STEP", i)
        step_fam.add_trigger("../../forecast/run_forecast:step ge " + str(i))
        step_fam.add_task("save")

    # Defines the triggers for the next cycle
    cycle_triggers = "./" + str(cycle) + " == complete"



#!/usr/bin/env python3.6
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import ecflow
import os       # for getenv
import sys
import shutil   # used to remove directory tree
import argparse # for argument parsing     

def sync_local(ci):
    ci.sync_local()
    return ci.get_defs() 

def get_tasks_to_run(defs,paths):
    print("Collating tasks to run in batch")
    node_path_list = []
    for path in paths:
        node = defs.find_abs_node(path)
        if not node:
            print("Could not find node",path,"in the definition")
            continue
        if isinstance(node,ecflow.Task):
            node_path_list.append(node.get_abs_node_path())
        elif isinstance(node,ecflow.NodeContainer):
            node_vec = node.get_all_nodes()
            for n in node_vec:
                if isinstance(n,ecflow.Task):
                    node_path_list.append(n.get_abs_node_path())
             
    uniqueList = []
    for path in node_path_list:
        if path not in uniqueList:
            uniqueList.append(path)
            print(" found task",path)
           
    return uniqueList 
    
def chunks(list_to_break, chunk_size):
    for i in range(0, len(list_to_break), chunk_size):
        yield list_to_break[i:i + chunk_size]
        
def run_chunk(ci,list_of_task_paths):
    defs = sync_local(ci)
    CL.suspend(list_of_task_paths);


if __name__ == "__main__":
    
    DESC = """ 
            example:
                python3 libs/pyext/samples/run_batches_with_delay.py -h polonius -p 4141 -b 4 -s 4 -n "/ecflow/cron /limit_basic"
                The list of paths can be obtained from the GUI, using:
                sh $WK/libs/pyext/samples/run_batches_with_delay.sh -h %ECF_HOST% -p %ECF_PORT% -b 10 -s 2 -n '<full_name>'
            """    
            
    print("####################################################################")
    print("Running ecflow version " + ecflow.Client().version()  + " debug build(" + str(ecflow.debug_build()) +")")
    if 'PYTHONPATH' in os.environ:
        print("PYTHONPATH: " + str(os.environ['PYTHONPATH'].split(os.pathsep)))
    print("sys.path:   " + str(sys.path))
    print("####################################################################")
 
    default_port = "3141"
    if "ECF_PORT" in os.environ:
         default_port = os.environ["ECF_PORT"]
 
    default_host = "localhost"
    if "ECF_HOST" in os.environ:
        default_host  = os.environ["ECF_HOST"]

    PARSER = argparse.ArgumentParser(description=DESC,  
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    PARSER.add_argument('--host', default=default_host,   
                        help="The name of the host machine, defaults to 'localhost'")
    PARSER.add_argument('--port', default=default_port,   
                        help="The port on the host, defaults to 3141")
    PARSER.add_argument('--batch_size', default=10, type=int,
                        help="The number of jobs to submit in batch")
    PARSER.add_argument('--sleep_between_batch', default=5, type=int,  
                        help="The number of seconds to sleep between the next batch")
    PARSER.add_argument('--paths',
                        help="This list of node path to execute. These could be suites/families.")
    ARGS = PARSER.parse_args()
    print(ARGS )  
    
    paths = ARGS.paths.split()
    print(paths)
    
    # ===========================================================================
    CL = ecflow.Client(ARGS.host, ARGS.port)
    try:
        print("check server " + ARGS.host + ":" + ARGS.port + " is running")
        CL.ping() 
        print("Server is running.")
         
        defs = sync_local(CL)
        node_path_list = get_tasks_to_run(defs,paths)
        
        list_of_chunks = (list(chunks(node_path_list,ARGS.batch_size)))
        for chunk in list_of_chunks:
            print(chunk)
            run_chunk(CL,chunk)
        
    except RuntimeError as ex:
        print("Error: " + str(ex))



#!/usr/bin/env python
""" ./ecflow3of5.py 63 /e_41r2/main/12/prod """
import ecflow
import os
import sys
import time
host = os.getenv("ECF_HOST", "localhost")
port = os.getenv("ECF_PORT", 31415)
client = ecflow.Client(host, port)
wait = False; wait = True
interval = 30
outof5 = int(sys.argv[1])
node_path = sys.argv[2]
def stop(msg, num): print msg; sys.exit(num)
while 1:
    tot = 0
    count = 0
    done = True
    client.sync_local()
    node = client.get_defs().find_abs_node(node_path)
    if node is None: stop("node not found!!!", 1)
    for item in node.nodes:
        count += 1
        status = "%s" % item.get_state()
        # print item.get_abs_node_path(), status, outof5, tot   
        if status == "complete":
            tot += 1
            if tot >= outof5: stop("# OK", 0)
        elif status == "aborted":
            pass
        else: done = False
    if count < outof5: stop("# Impossible: %d < %d" % (count, outof5), 1)
    if done: stop("# KO %d" % tot, 1)
    print "# still possible",
    if wait:
        print "...", tot, outof5, count
        time.sleep(interval);
    else: stop("", -1)


#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# code for testing addition and deletion 
#
import sys
import os
import copy
import unittest # for assertItemsEqual
import ecflow
import ecflow_test_util as Test

if __name__ == "__main__":
    
    Test.print_test_start(os.path.basename(__file__))
 
    #===========================================================================
    # Defs: add, delete and sort *USER* variables, use set(a).intersection(b) to compare lists
    #===========================================================================
    defs = ecflow.Defs()
    defs.add_variable("ZFRED", "/tmp/")
    defs.add_variable("YECF_URL_CMD", "${BROWSER:=firefox} -new-tab %ECF_URL_BASE%/%ECF_URL%")
    defs.add_variable("XECF_URL_BASE", "http://www.ecmwf.int")
    defs.add_variable("AECF_URL", "publications/manuals/sms")
    assert len(list(defs.user_variables)) == 4, "Expected *user* 4 variable"    
    
    # sort
    expected = ['AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED']
    actual = []
    defs.sort_attributes("variable");
    for v in defs.user_variables: actual.append(v.name())
    assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    expected = ['AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
    actual = []
    defs.add_variable("ZZ", "x")
    defs.sort_attributes(ecflow.AttrType.variable);
    for v in defs.user_variables: actual.append(v.name())
    assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    expected = ['AA', 'AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
    actual = []
    defs.add_variable("AA", "x")
    defs.sort_attributes("all");
    for v in defs.user_variables: actual.append(v.name())
    assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    expected = ['AA', 'AECF_URL', 'BB','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
    actual = []
    defs.add_variable("BB", "x")
    defs.sort_attributes(ecflow.AttrType.all );
    for v in defs.user_variables: actual.append(v.name())
    assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    
    
    defs.delete_variable("ZFRED");assert len(list(defs.user_variables)) == 6, "Expected 6 variables since we just delete ZFRED"
    defs.delete_variable("");     assert len(list(defs.user_variables)) == 0, "Expected 0 variables since we should have deleted all"
    
    a_dict = { "name":"value", "name2":"value2", "name3":"value3", "name4":"value4" }
    defs.add_variable(a_dict)
    assert len(list(defs.user_variables)) == 4, "Expected 4 variable"    
    defs.delete_variable("");         assert len(list(defs.user_variables)) == 0, "Expected 0 variable since we should have deleted all"

    # add a empty dictionary
    a_dict = {}
    defs.add_variable(a_dict)
    assert len(list(defs.user_variables)) == 0, "Expected zero variables"    
    
    a_dict = { "name":0, "name2":1, "name3":2, "name4":3 }
    defs.add_variable(a_dict)
    assert len(list(defs.user_variables)) == 4, "Expected 4 variable"    
    defs.delete_variable("");         assert len(list(defs.user_variables)) == 0, "Expected 0 variable since we should have deleted all"

    #===========================================================================
    # Suite: add,delete and sort variables
    #===========================================================================
    suite = ecflow.Suite("s1")
    suite.add_variable(ecflow.Variable("ECF_HOME", "/tmp/"))
    suite.add_variable("ZZZZZZ", "${BROWSER:=firefox} -new-tab %ECF_URL_BASE%/%ECF_URL%")
    suite.add_variable("YYYYY", "http://www.ecmwf.int")
    suite.add_variable("aaaa", "publications/manuals/sms")
    assert len(list(suite.variables)) == 4, "Expected 4 variable"    
    
    # sort
    expected = ['aaaa','ECF_HOME','YYYYY','ZZZZZZ']
    actual = []
    suite.sort_attributes("variable");
    suite.sort_attributes(ecflow.AttrType.variable);
    for v in suite.variables: actual.append(v.name())
    assert set(expected).intersection(actual), "Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    suite.delete_variable("ECF_HOME"); assert len(list(suite.variables)) == 3, "Expected 3 variable since we just delete ECF_HOME"
    suite.delete_variable("");         assert len(list(suite.variables)) == 0, "Expected 0 variable since we should have deleted all"
    
    a_dict = { "name":"value", "name2":"value2", "name3":"value3", "name4":"value4" }
    suite.add_variable(a_dict)
    assert len(list(suite.variables)) == 4, "Expected 4 variable"    
    suite.delete_variable("");         assert len(list(suite.variables)) == 0, "Expected 0 variable since we should have deleted all"

    a_dict = { "name":1, "name2":2, "name3":3, "name4":4 }
    suite.add_variable(a_dict)
    assert len(list(suite.variables)) == 4, "Expected 4 variable"    
    suite.delete_variable("");         assert len(list(suite.variables)) == 0, "Expected 0 variable since we should have deleted all"

    # add a empty dictionary
    a_dict = { }
    suite.add_variable(a_dict)
    assert len(list(suite.variables)) == 0, "Expected zero variables"    

    # adding dictionary items that are not strings,or ints in value ,should result in a type error
    expected_type_error = False
    try:
        a_bad_dict = { "name":"fred", "name2":14, "name3": list( 1,2,3), "name4":12 }
        suite.add_variable(a_bad_dict)
    except TypeError:
        expected_type_error = True
        
    assert expected_type_error, "Expected Type error"
    assert len(list(suite.variables)) == 0, "Expected 0 variable since we should have deleted all"
    
    suite.add_variable("ECF_URL_CMD", "test duplicates") 
    suite.add_variable("ECF_URL_CMD", "Expected warning")  # expect a warning message to standard out


    #===========================================================================
    # add,delete and sort limits
    #===========================================================================
    suite.add_limit(ecflow.Limit("zlimitName1", 10))
    suite.add_limit(ecflow.Limit("ylimitName2", 10))
    suite.add_limit("xlimitName3", 10)
    suite.add_limit("alimitName4", 10)
    assert len(list(suite.limits)) == 4, "Expected 4 Limits"
    
    # sort
    expected = ['alimitName4','xlimitName3','ylimitName2','zlimitName1']
    actual = []
    suite.sort_attributes(ecflow.AttrType.limit);
    suite.sort_attributes("limit");
    for v in suite.limits: actual.append(v.name())
    assert expected == actual, "Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)

    suite.delete_limit("zlimitName1"); assert len(list(suite.limits)) == 3, "Expected 3 limits since we just deleted one limitName1" 
    suite.delete_limit("");            assert len(list(suite.limits)) == 0, "Expected 0 limits since we just deleted all of them"

    # The following will fail, since the iterators are essentially read only, 
    # This is because we are using C++ vector iterators, hence we can't delete the vectors items, whilst traversing 
    #    for limit in suite.limits: suite.delete_limit(limit.name())
    # We can get round this by copy the limits first
    suite.add_limit(ecflow.Limit("limitName1", 10))
    suite.add_limit(ecflow.Limit("limitName2", 10))
    suite.add_limit("limitName3", 10)
    suite.add_limit("limitName4", 10)
    limit_names = []
    for limit in suite.limits:
        limit_names.append(limit.name())
    for limit in limit_names:
        suite.delete_limit(limit)
    assert len(list(suite.limits)) == 0, "Expected 0 Limits,since we just deleted all of them, via iteration"

    #===========================================================================
    # Test Limit and node paths, ECFLOW-518
    #===========================================================================
    the_limit = ecflow.Limit("limitName1", 10)
    assert the_limit.name() == "limitName1", "name not as expected"
    assert the_limit.value() == 0 ,"Expected limit value of 0"
    assert the_limit.limit() == 10 ,"Expected limit of 10"
    assert len(list(the_limit.node_paths())) == 0 ,"Expected nodes which have consumed a limit to be empty"
    
    the_limit.increment(1,"/path1") ;  assert the_limit.value() == 1 ,"Expected limit value of 1"
    the_limit.increment(1,"/path2") ;  assert the_limit.value() == 2 ,"Expected limit value of 2"
    the_limit.increment(1,"/path3") ;  assert the_limit.value() == 3 ,"Expected limit value of 3"
    the_limit.increment(1,"/path4") ;  assert the_limit.value() == 4 ,"Expected limit value of 4"
    for path in the_limit.node_paths(): print(path)
    assert len(list(the_limit.node_paths())) == 4 ,"expected 4 path"
    the_limit.decrement(1,"/path1") ; assert the_limit.value() == 3 ,"Expected limit value of 3"
    the_limit.decrement(1,"/path2") ; assert the_limit.value() == 2 ,"Expected limit value of 2"
    the_limit.decrement(1,"/path3") ; assert the_limit.value() == 1 ,"Expected limit value of 1"
    the_limit.decrement(1,"/path4") ; assert the_limit.value() == 0 ,"Expected limit value of 0"
    assert len(list(the_limit.node_paths())) == 0 ,"Expected nodes which have consumed a limit to be empty"

    the_limit.increment(1,"/path1") # add same path, should only consume one token
    the_limit.increment(1,"/path1")
    the_limit.increment(1,"/path1")
    the_limit.increment(1,"/path1")
    assert len(list(the_limit.node_paths())) == 1 ,"expected 1 path"
    assert the_limit.value() == 1 ,"Expected limit value of 1"

    #===========================================================================
    # add and delete inlimits
    #===========================================================================
    suite.add_inlimit(ecflow.InLimit("limitName1", "/s1/f1", 2))
    suite.add_inlimit(ecflow.InLimit("limitName2", "/s1/f1", 2, True))
    suite.add_inlimit("limitName3", "/s1/f1", 2)
    suite.add_inlimit("limitName4", "/s1/f1", 2)
    suite.add_inlimit("limitNameA", "/s1/f1", 2)
    suite.add_inlimit("limitNameA", "/s1/f1/a", 2)
    suite.add_inlimit("limitNameA", "/s1/f1/b", 2)
    assert len(list(suite.inlimits)) == 7, "Expected 7 inLimits"
    suite.delete_inlimit("limitName1"); assert len(list(suite.inlimits)) == 6, "Expected 6 inlimits since we just deleted one limitName1" 
    suite.delete_inlimit("/s1/f1:limitNameA"); assert len(list(suite.inlimits)) == 5, "Expected 5 inlimits since we just deleted one limitName1" 
    suite.delete_inlimit("/s1/f1/a:limitNameA"); assert len(list(suite.inlimits)) == 4, "Expected 4 inlimits since we just deleted one limitName1" 
    suite.delete_inlimit("/s1/f1/b:limitNameA"); assert len(list(suite.inlimits)) == 3, "Expected 3 inlimits since we just deleted one limitName1" 
    suite.delete_inlimit("");           assert len(list(suite.inlimits)) == 0, "Expected 0 inlimits since we just deleted all of them"

    #===============================================================================
    # add and delete triggers and complete
    #===============================================================================
    task = ecflow.Task("task")
    task.add_trigger("t2 == active")
    task.add_complete("t2 == complete")
    assert task.get_complete(), "Expected complete"
    assert task.get_trigger(), "Expected trigger"
    assert task.get_trigger().get_expression() == "t2 == active","add trigger failed"
    assert task.get_complete().get_expression() == "t2 == complete","add complete failed"
    task.delete_trigger();  assert not task.get_trigger(), "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"
    
    task.add_part_trigger(ecflow.PartExpression("t1 == complete"))
    task.add_part_trigger(ecflow.PartExpression("t2 == active", True))  #  for long and/or expressions, subsequent expr must be and/or
    task.add_part_complete(ecflow.PartExpression("t3 == complete"))
    task.add_part_complete(ecflow.PartExpression("t4 == active", False))  #  for long and/or expressions, subsequent expr must be and/or
    task.delete_trigger();  assert not task.get_trigger(), "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"

    task.add_part_trigger("t1 == complete")
    task.add_part_trigger("t2 == active", True)  #  for long and/or expressions, subsequent expr must be and/or
    task.add_part_complete("t3 == complete")
    task.add_part_complete("t4 == active", False)  #  for long and/or expressions, subsequent expr must be and/or
    task.delete_trigger();  assert not task.get_trigger(), "Expected no trigger"
    task.delete_complete(); assert not task.get_complete(), "Expected no complete"
    
    #===========================================================================
    # Add triggers using expressions
    #===========================================================================
    expr = ecflow.Expression("t1 == complete")
    task = ecflow.Task("task")
    task.add_trigger(expr)
    assert task.get_trigger().get_expression() == "t1 == complete","add trigger failed : " + task.get_trigger().get_expression()

    expr = ecflow.Expression("t1 == complete")
    expr.add(ecflow.PartExpression("t2 == complete", True))
    expr.add(ecflow.PartExpression("t3 == complete", True))
    task = ecflow.Task("task")
    task.add_trigger(expr)
    assert task.get_trigger().get_expression() == "t1 == complete AND t2 == complete AND t3 == complete","add trigger failed : " + task.get_trigger().get_expression()

    #===========================================================================
    # add,delete,find events
    #===========================================================================
    task.add_event(ecflow.Event(1)); 
    task.add_event(2)                              
    task.add_event(ecflow.Event(10, "Eventname"))  
    task.add_event(10, "Eventname2")               
    task.add_event("fred")                         
    
    # sort
    expected = ['1','2','Eventname','Eventname2','fred']
    actual = []
    task.sort_attributes("event");
    task.sort_attributes(ecflow.AttrType.event);
    for v in task.events: actual.append(v.name_or_number())
    assert expected == actual, "Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
     
    # test find
    event = task.find_event("EVENT")
    assert(event.empty()),"Expected to not to find event"
    assert(event.name() == ""),"Expected to not to find event, number is maximum int"
    assert(event.value() == 0),"Expected to not to find event"

    event = task.find_event("1"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 1), "Expected to find event 1"
    assert(event.name() == ""), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task.find_event("2"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 2), "Expected to find event 1"
    assert(event.name() == ""), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task.find_event("10"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.number() == 10), "Expected to find event 10"
    assert(event.name() == "Eventname"), "Expected name to be empty"
    assert(event.value() == 0),"Expected to not to find event"

    event = task.find_event("fred"); 
    assert(not event.empty()),"Expected to find event"
    assert(event.name() == "fred"), "Expected name to be empty, when name defind an not number, number is max_int"
    assert(event.value() == 0),"Expected to not to find event"

    for e in task.events:     print(str(e)," # value: ",str(e.value()))
    a_dict = {}
    for e in task.events:
        if e.name() != "": a_dict[e.name()] = e.value()
        else:              a_dict[e.number()] = e.value()
    print(a_dict)

    assert len(list(task.events)) == 5, "Expected 5 Events"
    task.delete_event("1");         assert len(list(task.events)) == 4, "Expected 4 Events"
    task.delete_event("Eventname"); assert len(list(task.events)) == 3, "Expected 3 Events"
    task.delete_event("");          assert len(list(task.events)) == 0, "Expected 0 Events"


    #===========================================================================
    # add and delete meter
    #===========================================================================
    task.add_meter(ecflow.Meter("zzzz", 0, 100, 50))
    task.add_meter(ecflow.Meter("yyyy", 0, 100))
    task.add_meter("bbbb", 0, 100, 50)
    task.add_meter("aaaa", 0, 100)
    assert len(list(task.meters)) == 4, "Expected 4 Meters"
    
    # sort
    expected = ['aaaa','bbbb','yyyy','zzzz']
    actual = []
    task.sort_attributes(ecflow.AttrType.meter);
    task.sort_attributes("meter");
    for v in task.meters: actual.append(v.name())
    assert expected == actual, "Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
 
    task.delete_meter("zzzz"); assert len(list(task.meters)) == 3, "Expected 3 Meters"
    task.delete_meter("yyyy"); assert len(list(task.meters)) == 2, "Expected 2 Meters"
    task.delete_meter("");     assert len(list(task.meters)) == 0, "Expected 0 Meters"

    #===========================================================================
    # add and delete queue
    #===========================================================================
    queue_items = [ "001", "002"]
    task.add_queue(ecflow.Queue("queue",queue_items))
    task.add_queue("queue1",queue_items)
    task.add_queue("queue2",queue_items)
    task.add_queue("queue3",queue_items)
    assert len(list(task.queues)) == 4, "Expected 4 Queues"
    task.delete_queue("queue3"); assert len(list(task.queues)) == 3, "Expected 3 Queues"
    task.delete_queue("queue2"); assert len(list(task.queues)) == 2, "Expected 2 Queues"
    task.delete_queue("");       assert len(list(task.queues)) == 0, "Expected 0 Queues"

    #===========================================================================
    # add and delete, find generic. These are arbitary attributes for future func.
    #===========================================================================
    generic_items = [ "001", "002"]
    task.add_generic(ecflow.Generic("gen",generic_items ))
    task.add_generic("gen1",generic_items )
    task.add_generic("gen2",generic_items )
    task.add_generic("gen3",generic_items )
    assert len(list(task.generics)) == 4, "Expected 4 Generics"
    gen = task.find_generic("gen3")
    assert not gen.empty()," find failed"
    assert gen.name() == "gen3", "expected for find gen3"
    
    task.delete_generic("gen1"); assert len(list(task.generics)) == 3, "Expected 3 Generics"
    task.delete_generic("gen2"); assert len(list(task.generics)) == 2, "Expected 2 Generics"
    task.delete_generic("");     assert len(list(task.generics)) == 0, "Expected 0 Generics"

    #===========================================================================
    # add and delete label
    #===========================================================================
    task.add_label(ecflow.Label("labela", "value"))
    task.add_label(ecflow.Label("labelb", "value"))
    task.add_label("labelc", "value")
    task.add_label("labeld", "value")
    assert len(list(task.labels)) == 4, "Expected 4 labels"
    
    # sort
    expected = ['labela','labelb','labelc','labeld']
    actual = []
    task.sort_attributes("label");
    task.sort_attributes(ecflow.AttrType.label);
    for v in task.labels: actual.append(v.name())
    assert expected == actual, "Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
    
    task.delete_label("labela"); assert len(list(task.labels)) == 3, "Expected 3 Labels"
    task.delete_label("labelb"); assert len(list(task.labels)) == 2, "Expected 2 Labels"
    task.delete_label("");       assert len(list(task.labels)) == 0, "Expected 0 Labels"

    #===========================================================================
    # add delete Repeat
    #===========================================================================
    task.add_repeat(ecflow.RepeatInteger("integer", 0, 100, 2))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    task.add_repeat(ecflow.RepeatEnumerated("enum", ["red", "green", "blue" ]))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat(ecflow.RepeatDate("date", 20100111, 20100115, 2))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat(ecflow.RepeatDateTime("datetime", "20100111T000000", "20100115T000000", "48:00:00"))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat(ecflow.RepeatDateList("date",[20100111, 20100115]))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
     
    task.add_repeat(ecflow.RepeatString("string", ["a", "b", "c" ]))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"

    task.add_repeat(ecflow.RepeatDay(1))
    repeat = task.get_repeat(); assert not repeat.empty(), "Expected repeat"
    task.delete_repeat()      
    repeat = task.get_repeat(); assert repeat.empty(), "Expected no repeat"
    
    
    #===========================================================================
    # create a time series, used for adding time and today
    #===========================================================================
    start = ecflow.TimeSlot(0, 0)
    finish = ecflow.TimeSlot(23, 0)
    incr = ecflow.TimeSlot(0, 30)
    time_series = ecflow.TimeSeries(start, finish, incr, True)

    # Add and delete today
    # NOTE: **********************************************************************
    # Do *NOT* delete attributes using iterator traversal
    # The iterators *are* bound to the c++ iterators hence if we delete
    # over the traversal, we can corrupt the vector, leading to undefined behavour.  
    # Hence we **cannot** delete more than once over a traversal
    # Soln 1: take a copy
    # *****************************************************************************
    # task.add_today( Today( 0,10 ))
    # task.add_today( Today( 0,59, True ))
    # task.add_today( Today(TimeSlot(20,10)) )
    # task.add_today( Today(TimeSlot(20,20),False)) 
    # for today in task.todays: 
    #    task.delete_today(today) # will corrupt C++ vector
  
    task.add_today("00:30")
    task.add_today("+00:30")
    task.add_today("+00:30 20:00 01:00")
    task.add_today(ecflow.Today(time_series))
    task.add_today(ecflow.Today("+00:30"))
    task.add_today(ecflow.Today("+00:30 20:00 01:00"))
    task.add_today(ecflow.Today(0, 10))
    task.add_today(0, 59, True)
    task.add_today(ecflow.Today(ecflow.TimeSlot(20, 10)))
    task.add_today(ecflow.Today(ecflow.TimeSlot(20, 20), False)) 
    assert len(list(task.todays)) == 10, "Expected 8 todays"
    vec_copy = []
    for today in task.todays: vec_copy.append(today)
    for today in vec_copy: task.delete_today(today)
    assert len(list(task.todays)) == 0, "Expected 0 todays"
        
    #===========================================================================
    # add and delete time
    #===========================================================================
    task.add_time("00:30")
    task.add_time("+00:30")
    task.add_time("+00:30 20:00 01:00")
    task.add_time(ecflow.Time(time_series))
    task.add_time(ecflow.Time(0, 10))
    task.add_time(ecflow.Time("+00:30"))
    task.add_time(ecflow.Time("+00:30 20:00 01:00"))
    task.add_time(0, 59, True)
    task.add_time(ecflow.Time(ecflow.TimeSlot(20, 10)))
    task.add_time(ecflow.Time(ecflow.TimeSlot(20, 20), False)) 
    assert len(list(task.times)) == 10, "Expected 8 times"
    vec_copy = []
    for time in task.times: vec_copy.append(time)
    for time in vec_copy: task.delete_time(time)
    assert len(list(task.todays)) == 0, "Expected 0 todays"

    #===========================================================================
    # add and delete date
    #===========================================================================
    for i in [ 1, 2, 4, 8, 16 ] :
        task.add_date(i, 0, 0)
    task.add_date(ecflow.Date(1, 1, 2010))
    task.add_date(1, 1, 2010)       # duplicate
    task.add_date(ecflow.Date(2, 1, 2010))
    task.add_date(ecflow.Date(3, 1, 2010))
    task.add_date(ecflow.Date(4, 1, 2010)) 
    assert len(list(task.dates)) == 10, "Expected 10 dates but found " + str(len(list(task.dates)))
    vec_copy = []
    for attr in task.dates: vec_copy.append(attr)
    for attr in vec_copy: task.delete_date(attr)
    assert len(list(task.dates)) == 0, "Expected 0 dates"

    #===========================================================================
    # add and delete day
    #===========================================================================
    task.add_day(ecflow.Day(ecflow.Days.sunday))
    task.add_day(ecflow.Days.monday)
    task.add_day(ecflow.Days.tuesday)  # duplicate ?
    task.add_day(ecflow.Day("tuesday"))   
    task.add_day("sunday")     
    assert len(list(task.days)) == 5, "Expected 5 days"
    vec_copy = []
    for attr in task.days: vec_copy.append(attr)
    for attr in vec_copy: task.delete_day(attr)
    assert len(list(task.days)) == 0, "Expected 0 days"
    
    #===========================================================================
    # add and delete crons
    #===========================================================================
    cron = ecflow.Cron()
    start = ecflow.TimeSlot(23 , 0)
    ts = ecflow.TimeSeries(start,True)  # True means relative to suite start
    cron.set_time_series(ts)
    
    cron = ecflow.Cron()
    cron.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron.set_days_of_month([1, 2, 3, 4, 5, 6 ])
    cron.set_months([1, 2, 3, 4, 5, 6])
    start = ecflow.TimeSlot(0 , 0)
    finish = ecflow.TimeSlot(23, 0)
    incr = ecflow.TimeSlot(0, 30)
    ts = ecflow.TimeSeries(start, finish, incr, True)  # True means relative to suite start
    cron.set_time_series(ts)

    cron0 = ecflow.Cron()
    cron0.set_week_days([0, 1, 2, 3, 4, 5 ])
    cron0.set_time_series(1, 30) # default relative = false, added in release 4.0.7

    cron1 = ecflow.Cron()
    cron1.set_week_days([0, 1, 2, 3, 4, 5, 6 ])
    cron1.set_time_series(1, 30, True)
    
    cron2 = ecflow.Cron()
    cron2.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron2.set_time_series("00:30 01:30 00:01")
    
    cron3 = ecflow.Cron()
    cron3.set_week_days([0, 1, 2, 3, 4, 5, 6])
    cron3.set_time_series("+00:30")

    task.add_cron(cron)
    task.add_cron(cron1)
    task.add_cron(cron2)
    task.add_cron(cron3)
    assert len(list(task.crons)) == 4, "Expected 4 crons"
    vec_copy = []
    for attr in task.crons: vec_copy.append(attr)
    for attr in vec_copy: task.delete_cron(attr)
    assert len(list(task.crons)) == 0, "Expected 0 crons"  
    
    #===========================================================================
    # add autocancel
    #===========================================================================
    print("test add autoCancel")
    t1 = ecflow.Task("t1")
    assert t1.get_autocancel() is None, " Expected no autocancel"
    t1.add_autocancel(3)                       # 3 days
    assert t1.get_autocancel() is not None, " Expected autocancel"
    print(str(t1.get_autocancel()))

    t3 = ecflow.Task("t3")
    t3.add_autocancel(20, 10, True)              # hour,minutes,relative
    print(str(t3.get_autocancel()))
    
    t4 = ecflow.Task("t4")
    t4.add_autocancel(ecflow.TimeSlot(10, 10), True)    # hour,minutes,relative
    print(str(t4.get_autocancel()))

    t5 = ecflow.Task("t5")
    t5.add_autocancel(ecflow.Autocancel(1, 10, True))   # hour,minutes,relative
    print(str(t5.get_autocancel()))
    
    #===========================================================================
    # add autoarchive
    #===========================================================================
    print("test add autoarchive")
    f1 = ecflow.Family("f1")
    assert f1.get_autoarchive() is None, " Expected no autoarchive"
    f1.add_autoarchive(3)                       # 3 days
    assert f1.get_autoarchive() is not None, " Expected autoarchive"
    print(str(f1.get_autoarchive()))
 
    f2 = ecflow.Family("f2")
    assert f2.get_autoarchive() is None, " Expected no autoarchive"
    f2.add_autoarchive(3,True)                       # 3 days
    assert f2.get_autoarchive() is not None, " Expected autoarchive"
    print(str(f2.get_autoarchive()))

    f3 = ecflow.Family("f3")
    f3.add_autoarchive(20, 10, True)              # hour,minutes,relative
    print(str(f3.get_autoarchive()))
    
    f3_1 = ecflow.Family("f3_1")
    f3_1.add_autoarchive(20, 10, True,True)          # hour,minutes,relative,idle
    print(str(f3_1.get_autoarchive()))

    f4 = ecflow.Family("f4")
    f4.add_autoarchive(ecflow.TimeSlot(10, 10), True)    # hour,minutes,relative
    print(str(f4.get_autoarchive()))

    f4_1 = ecflow.Family("f4_1")
    f4_1.add_autoarchive(ecflow.TimeSlot(10, 10), True,True)    # hour,minutes,relative,idle
    print(str(f4_1.get_autoarchive()))

    f4_2 = ecflow.Family("f4_2")
    assert f4_2.get_autoarchive() is None, " Expected no autoarchive"
    f4_2.add_autoarchive(ecflow.Autoarchive(10,True))    # Days,idle
    assert f4_2.get_autoarchive() is not None, " Expected autoarchive"
    print(str(f4_2.get_autoarchive()))

    #===========================================================================
    # autocancel
    #===========================================================================
    f5 = ecflow.Family("f5")
    f5.add_autocancel(ecflow.Autocancel(1, 10, True))   # hour,minutes,relative
    print(str(f5.get_autocancel()))
    
    #===========================================================================
    # add autorestore
    #===========================================================================
    print("test add autorestore")
    t1 = ecflow.Task("f1")
    assert t1.get_autorestore() is None, " Expected no autorestore"
    t1.add_autorestore(["/s1/f2"])                       
    assert t1.get_autorestore() is not None, " Expected autorestore"
    print(str(t1.get_autorestore()))

    f5 = ecflow.Family("f5")
    assert f5.get_autorestore() is None, " Expected no autorestore"
    f5.add_autorestore(["/s1/f2","/s2","/s3"])                       
    assert f5.get_autorestore() is not None, " Expected autorestore"
    print(str(f5.get_autoarchive()))
    
    f6 = ecflow.Family("f6")
    assert f6.get_autorestore() is None, " Expected no autorestore"
    f6.add_autorestore(ecflow.Autorestore(["/s1/f2","/s2","/s3"]))                       
    assert f6.get_autorestore() is not None, " Expected autorestore"
    print(str(f6.get_autoarchive()))
    
    #===========================================================================
    # add late
    #===========================================================================
    late = ecflow.Late()
    late.submitted(ecflow.TimeSlot(20, 10))
    late.active(ecflow.TimeSlot(20, 10))
    late.complete(ecflow.TimeSlot(20, 10), True)
    task.add_late(late)
    
    late = ecflow.Late()
    late.submitted(20, 10)
    late.active(20, 10)
    late.complete(20, 10, True)
    t1.add_late(late)
    
    
    #===========================================================================
    # add defstatus, last one set takes effect
    #===========================================================================
    task.add_defstatus(ecflow.DState.complete)
    assert task.get_defstatus() == ecflow.DState.complete
    task.add_defstatus(ecflow.DState.queued)
    assert task.get_defstatus() == ecflow.DState.queued
    task.add_defstatus(ecflow.DState.aborted)
    assert task.get_defstatus() == ecflow.DState.aborted
    task.add_defstatus(ecflow.DState.submitted)
    assert task.get_defstatus() == ecflow.DState.submitted
    task.add_defstatus(ecflow.Defstatus(ecflow.DState.suspended ));
    assert task.get_defstatus() == ecflow.DState.suspended
    task.add_defstatus(ecflow.Defstatus("active"))  
    assert task.get_defstatus() == ecflow.DState.active

    # the state should be unknown until the suite is begun
    assert task.get_state() == ecflow.State.unknown
    assert task.get_dstate() == ecflow.DState.unknown
 
    #===========================================================================
    # add clock
    #===========================================================================
    clock = ecflow.Clock(1, 1, 2010, False)     # day,month, year, hybrid
    clock.set_gain(1, 10, True)                 # True means positive gain
    suite = ecflow.Suite("suite")
    suite.add_clock(clock)
    
    clock = ecflow.Clock(1, 1, 2011, True)       # day,month, year, hybrid
    clock.set_gain_in_seconds(12, True)
    s1 = ecflow.Suite("s1")
    s1.add_clock(clock)
    
    print("#===========================================================================")
    print("# get clock")
    print("#===========================================================================")
    s0 = ecflow.Suite("s0")
    assert s0.get_clock() is None, "Expected no clock"
    
    s0.add_clock(ecflow.Clock(1, 1, 2010, False))
    assert s0.get_clock() is not None, "Expected clock"

    #===========================================================================
    # end clock, used in simulator only, not persisted
    #===========================================================================
    clock = ecflow.Clock(1, 1, 2010, False)     # day,month, year, hybrid
    clock.set_gain(1, 10, True)                 # True means positive gain
    suite = ecflow.Suite("suite")
    suite.add_end_clock(clock)
    
    clock = ecflow.Clock(1, 1, 2011, True)       # day,month, year, hybrid
    clock.set_gain_in_seconds(12, True)
    s1 = ecflow.Suite("s1")
    s1.add_end_clock(clock)
    
    print("#===========================================================================")
    print("# get end clock")
    print("#===========================================================================")
    s0 = ecflow.Suite("s0")
    assert s0.get_end_clock() is None, "Expected no end clock"
    
    s0.add_end_clock(ecflow.Clock(1, 1, 2010, False))
    assert s0.get_end_clock() is not None, "Expected end clock"
    #===========================================================================
    # Add zombie. Note we can *NOT* add two zombie attributes of the same ZombieType
    #===========================================================================
    zombie_life_time_in_server = 800
    child_list = [ ecflow.ChildCmdType.init, ecflow.ChildCmdType.event, ecflow.ChildCmdType.meter, ecflow.ChildCmdType.label, ecflow.ChildCmdType.wait, ecflow.ChildCmdType.abort, ecflow.ChildCmdType.complete ]
    zombie_type_list = [ ecflow.ZombieType.ecf, ecflow.ZombieType.ecf_pid, ecflow.ZombieType.ecf_pid_passwd, ecflow.ZombieType.ecf_passwd, ecflow.ZombieType.user, ecflow.ZombieType.path ]
    for zombie_type in zombie_type_list:
        zombie_attr = ecflow.ZombieAttr(zombie_type, child_list, ecflow.ZombieUserActionType.block, zombie_life_time_in_server)
        s1.add_zombie(zombie_attr)
    assert len(list(s1.zombies)) == 6, "Expected 6 zombie attributes but found " + str(len(list(s1.zombies)))
    
    # delete all the zombies
    s1.delete_zombie("")
    assert len(list(s1.zombies)) == 0, "Expected zero zombie attributes but found " + str(len(list(s1.zombies)))
    
    # add with zombie_life_time_in_server not set, this is optional
    for zombie_type in zombie_type_list:
        zombie_attr = ecflow.ZombieAttr(zombie_type, child_list, ecflow.ZombieUserActionType.block)
        s1.add_zombie(zombie_attr)
    assert len(list(s1.zombies)) == 6, "Expected 6 zombie attributes but found " + str(len(list(s1.zombies)))

    s1.delete_zombie("")
    assert len(list(s1.zombies)) == 0, "Expected zero zombie attributes but found " + str(len(list(s1.zombies)))

    # repeat the the test with empty child list. Empty child list means apply to all child commands
    child_list = []
    for zombie_type in zombie_type_list:
        zombie_attr = ecflow.ZombieAttr(zombie_type, child_list, ecflow.ZombieUserActionType.block, zombie_life_time_in_server)
        s1.add_zombie(zombie_attr)
        
        # test delete of specific zombie attribute
        assert len(list(s1.zombies)) == 1, "Expected 1 zombie attributes but found " + str(len(list(s1.zombies)))
        s1.delete_zombie(zombie_type)
        assert len(list(s1.zombies)) == 0, "Expected 0 zombie attributes but found " + str(len(list(s1.zombies)))


    print("All Tests pass")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# code for testing with statement
# In our case it provides the ability to indent

from ecflow import Defs, Suite, Task, Family, Client, Variable, Edit, debug_build
import ecflow_test_util as Test
import sys,os
    
if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
 
    # sys.version_info is a tuple containing (major,minor,micro,releaselevel,serial)
    version = sys.version_info;
    if (sys.version_info < (2, 7)):
        print("This test only run with python version 2.7 or greater, but found : " + str(version))
        sys.exit(0)
    
    # Add with context manager, in our case this only provides ability to indent
    with Defs() as defs:
        with defs.add_suite("s2") as s2:
            with s2.add_family("f1") as f1:
                with f1.add_task("t1") as t1:
                    assert len(f1) == 1,"Expected 1 task"
            s2.add_family("f2")
            s2.add_task("t1")
        assert len(s2) == 3,"Expected 2 families and one task suite"
    assert len(defs) == 1,"Expected 1 suite"
#    print defs
 
    with Suite("s2") as s2:
        s2.add_family("f1")
        s2.add_task("t1")
        assert len(s2) == 2,"Expected 2 nodes family and task"
        
    with Family("f1") as f1:
        f1.add_family("f2")
        f1.add_task("t1")
        assert len(s2) == 2,"Expected 2 nodes family and task"
        
        
    #
    # Add Nodes functional way
    #
    defs1 = Defs()
    defs1.add_suite("s1").add_task("t1").add_variable("var","v")
    defs1.add_suite("s2").add_family("f1").add_task("t1").add_variable("var","v")
    defs1.add_suite("s3").add_family("f1").add_family("f2").add_task("t1").add_variable("var","v")
    
    # add node using with, they should compare
    with Defs() as defs2:
        with defs2.add_suite("s1") as s1:
            t1 = s1.add_task("t1")
            t1.add_variable("var","v")
        with defs2.add_suite("s2") as s2:
            with s2.add_family("f1") as f1:
                with f1.add_task("t1") as t1:
                    t1.add_variable("var","v")
        with defs2.add_suite("s3") as s3:
            with s3.add_family("f1") as f1:
                with f1.add_family("f2") as f2: 
                    with f2.add_task("t1") as t1: 
                        t1.add_variable("var","v")
    
    assert defs1 == defs2,"expected defs to be the same"     
    
    # add node using with, they should compare
    with Defs() as defs3:
        with defs3.add_suite("s1") as s1:
            s1 + Task('t1',var='v')
        with defs3.add_suite("s2") as s2:
            with s2.add_family("f1") as f1:
                with f1.add_task("t1") as t1:
                    t1 += Edit(var='v')
        with defs3.add_suite("s3") as s3:
            with s3.add_family("f1") as f1:
                with f1.add_family("f2") as f2: 
                    with f2.add_task("t1") as t1: 
                        t1 += Edit(var='v')
    
    assert defs1 == defs3,"expected defs to be the same"           
        
    print("All tests pass")


#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import unittest 
from ecflow import *

class Test_sort(unittest.TestCase):

    def get_object_list(self,obj,name_list):
        if obj == 'Meter': 
            return [ Meter(name,0,100) for name in name_list ]
        elif obj == 'Label':
            return [ Label(name,'value') for name in name_list ]
        elif obj == 'Variable':
            return [ Variable(name,'value') for name in name_list ]
        elif obj == 'Limit':
            return [ Limit(name,10) for name in name_list ]
        else:
            return [ eval(obj)(name) for name in name_list ]

    def test_sort_objects(self):
        names = [ 'zz',  'ww', 'aa']
        object_list = [ 'Task','Family','Suite','Event','Meter','Label','Variable','Limit']
        for obj in object_list:
            self.assertEqual(sorted(self.get_object_list(obj,names)), self.get_object_list(obj,sorted(names)), " sort of " + obj + " failed")
#             print(obj,'-----------------')
#             for item in sorted(self.get_object_list(obj,names)):
#                 print(item)
                
                
class Test_sort_api(unittest.TestCase):
    def test_sort_attributes_api(self):
        nodes = [ Defs(), Suite('s'), Family('f'), Task('t')]
        sort_attr = ['event','meter','label','variable','all']
        for attr in sort_attr:
            for node in nodes:
                node.sort_attributes(attr)
                node.sort_attributes(attr,True)
                node.sort_attributes(attr,True,['/path','/path2'])

        sort_attr = [AttrType.event, AttrType.meter, AttrType.label,AttrType.variable, AttrType.all  ]
        for attr in sort_attr:
            for node in nodes:
                node.sort_attributes(attr)
                node.sort_attributes(attr,True)
                node.sort_attributes(attr,True,['/path','/path2'])

# see ECFLOW-1642
class Test_ECFLOW_1642(unittest.TestCase):
    def get_events(self):
        return [Event(name) for name in
            "foo buzz bar google zulu mason".split()]

    def sorted_events(self):
        return [Event(name) for name in
            "bar buzz foo google mason zulu".split()]

    def expected_def(self):
        defs = Defs(Suite("mysuite",
                Family("fam1",self.get_events()),
                Family("fam2",self.sorted_events(),
                   Task("task1",self.sorted_events()),
                ),
                Task("task2",self.sorted_events()),
                self.sorted_events()))
        return defs

    def defs_to_sort(self):
        defs = Defs(Suite("mysuite",
                Family("fam1",self.get_events()),
                Family("fam2",self.get_events(),
                   Task("task1",self.get_events()),
                ),
                Task("task2",self.get_events()),
                self.get_events()))
        return defs
    
    def test_no_sort(self):
        defs = self.defs_to_sort()
        defs.sort_attributes('event',True,['/mysuite/fam1'])
        self.assertEqual(defs,self.expected_def(),"Expected\n" + str(self.expected_def()) + "but found\n" + str(defs))

    def test_no_sort2(self):
        defs = self.defs_to_sort()
        defs.sort_attributes(AttrType.event,True,['/mysuite/fam1'])
        self.assertEqual(defs,self.expected_def(),"Expected\n" + str(self.expected_def()) + "but found\n" + str(defs))
    
class Test_defs_sort(unittest.TestCase):
    def test_defs_sort_attributes(self):
        defs = Defs()
        defs.add_variable("ZFRED", "/tmp/")
        defs.add_variable("YECF_URL_CMD", "${BROWSER:=firefox} -new-tab %ECF_URL_BASE%/%ECF_URL%")
        defs.add_variable("XECF_URL_BASE", "http://www.ecmwf.int")
        defs.add_variable("AECF_URL", "publications/manuals/sms")
        assert len(list(defs.user_variables)) == 4, "Expected *user* 4 variable"    
     
        # sort
        expected = ['AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED']
        actual = []
        defs.sort_attributes("variable");
        for v in defs.user_variables: actual.append(v.name())
        assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
 
        expected = ['AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
        actual = []
        defs.add_variable("ZZ", "x")
        defs.sort_attributes(AttrType.variable);
        for v in defs.user_variables: actual.append(v.name())
        assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
     
        expected = ['AA', 'AECF_URL','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
        actual = []
        defs.add_variable("AA", "x")
        defs.sort_attributes("all");
        for v in defs.user_variables: actual.append(v.name())
        assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
     
        expected = ['AA', 'AECF_URL', 'BB','XECF_URL_BASE','YECF_URL_CMD','ZFRED','ZZ']
        actual = []
        defs.add_variable("BB", "x")
        defs.sort_attributes(AttrType.all );
        for v in defs.user_variables: actual.append(v.name())
        assert actual == expected,"Attributes not sorted, expected:" + str(expected) + " but found:" + str(actual)
 
if __name__ == "__main__":
    unittest.main()
    print("All Tests pass")


import os
from ecflow import (
    Defs,
    Suite,
    Family,
    Task,
    Edit,
    Trigger,
    Complete,
    Event,
    Meter,
    Time,
    Day,
    Date,
)

print("Creating suite definition")
home = os.path.join(os.getenv("HOME"), "course")
defs = Defs(
    Suite(
        "test",
        Edit(ECF_INCLUDE=home, ECF_HOME=home),
        Family(
            "f1",
            Edit(SLEEP=20),
            Task("t1", Meter("progress", 1, 100, 90)),
            Task("t2", Trigger("t1 == complete"), Event("a"), Event("b")),
            Task("t3", Trigger("t2:a")),
            Task("t4", Trigger("t2 == complete"), Complete("t2:b")),
            Task("t5", Trigger("t1:progress ge 30")),
            Task("t6", Trigger("t1:progress ge 60")),
            Task("t7", Trigger("t1:progress ge 90")),
        ),
        Family(
            "f2",
            Edit(SLEEP=20),
            Task("t1", Time("00:30 23:30 00:30")),
            Task("t2", Day("sunday")),
            Task("t3", Date("1.*.*"), Time("12:00")),
            Task("t4", Time("+00:02")),
            Task("t5", Time("00:02")),
        ),
    )
)
print(defs)

print("Checking job creation: .ecf -> .job0")
print(defs.check_job_creation())

print("Checking trigger expressions")
assert len(defs.check()) == 0, defs.check()

print("Saving definition to file 'test.def'")
defs.save_as_defs("test.def")



#!/usr/bin/env python2.7
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# 
# Handle bugs, produced by old versions of ecflow_client --migrate
#   - Mostly to do with handling newlines
#

import os          # for getenv
import shutil      # used to remove directory tree
#import argparse   # requires python 2.7
import getopt, sys # for argument parsing  

class Migrator(object):
    def __init__(self, list_of_input_lines):
        self.list_of_input_lines =  list_of_input_lines
        self.list_of_output_lines = []
        
    def _migration_hook(self,default_version_number):
        """Derived suite should override this function to perform the migration
        """
        return False;
    
    def output_lines(self):
        return self.list_of_output_lines
    
    def migrate(self, default_version_number):
        """Template/skeleton function, which will call the hooks
        """
        version = self.version_number_as_integer(default_version_number) 
        if version >= default_version_number:
            self.list_of_output_lines = self.list_of_input_lines
            return False

        self.list_of_output_lines = [] 
        migrated = self._migration_hook(default_version_number)
        if migrated : 
            self.list_of_input_lines = self.list_of_output_lines   # preserve this migration
            print("Applying ", type(self).__name__)  
        else:
            self.list_of_output_lines = self.list_of_input_lines   
        return migrated         

    def version_number_as_integer(self,default_version):
        """Determine the version number used in ecflow_client --migrate
           This should be in the very first line
           # 3.1.8
        """
        if len(self.list_of_input_lines) > 0:
            tokens = self.list_of_input_lines[0].split()
            if len(tokens) > 1:
                # print(tokens)
                version = (int)(tokens[1].replace(".",""))
                # print("******************* found version " , str(version))
                return version;
        return default_version
    
    
class MigrateForTaskAbort(Migrator):
    """Fix bug where the task abort states(i.e a string) will contain new lines and hence
       spill over the next line, and affecting the parser. This will move the abort string
       into the same line, and remove the interleaving new lines
    """
    def __init__(self, list_of_input_lines):
        Migrator.__init__(self, list_of_input_lines)
 
    def _migration_hook(self,default_version_number):
        """
        The correct format is: where abort state is on the same line
        task task # passwd:jxX0gIbR abort<: reason for abort >abort:
        ensure_start_abort_and_end_abort_on_sameline
        """
        migrated = False
        start_line_append = ""
        for line in self.list_of_input_lines:
            # process line
            end_abort_pos = line.find(">abort")
            task_pos = line.find("task")
            if task_pos != -1:
                start_abort_pos = line.find("abort<:")
                if start_abort_pos != -1 and end_abort_pos != -1:
                    # start and end found on same line. This is the correct format
                    start_line_append = ""
                
                if start_abort_pos != -1 and end_abort_pos == -1:
                    # print("start abort found with no end abort")
                    start_line_append = line
                    start_line_append = start_line_append.rstrip('\n')
                    start_line_append += " "
                    migrated = True
                    continue
    
            if len(start_line_append) > 0:
                start_line_append += line
                if end_abort_pos != -1:
                    self.list_of_output_lines.append(start_line_append) # preserve'\n' for this case                   
                    # print(start_line_append)
                    start_line_append = "";
                else:
                    start_line_append = start_line_append.rstrip('\n')
                    start_line_append += " "
    
                continue;
                  
            self.list_of_output_lines.append(line)
           
        return migrated  
    
    
class MigrateForLabel(Migrator):
    """Fix bug where label value has new lines, and hence spills over to next line
       This will remove place newline back onto the same line
    """
    def __init__(self, list_of_input_lines):
        Migrator.__init__(self, list_of_input_lines)
 
    def _migration_hook(self,default_version_number):
        """
        The correct format is where label in on the same line
            label name "value" # "new value"
        However if the value or new value has new line it can mess up the parsing
            label info "have you? set START_LOBSVR=1 in the task, once, to launch the logserver,
    setup ssh login
    created remote /tmp/map/cray"
        """
        #count = 0;
        no_of_label_quotes = 0
        migrated = False
        start_line_append = ""
        for line in self.list_of_input_lines:
            #count =  count + 1
            #print(str(count) , ": " , line)
            #if count == 383119:
            #    print("debug mee")
            if len(start_line_append) > 0:
                no_of_label_quotes += line.count('"')
                start_line_append += line
                if no_of_label_quotes % 2 == 0:
                    # even number of quotes
                    self.list_of_output_lines.append(start_line_append) # preserve last '\n'  
                    start_line_append = ""
                    no_of_label_quotes = 0 
                else:
                    start_line_append = start_line_append.rstrip('\n')
                continue            
            else:
            
                # process line
                tokens = line.split()
                if len(tokens) >0 and tokens[0] == "label":
                    no_of_label_quotes = line.count('"')
                    if no_of_label_quotes % 2 != 0 : 
                        start_line_append = line
                        start_line_append = start_line_append.rstrip('\n')
                        migrated = True
                        continue
                  
            self.list_of_output_lines.append(line)
           
        return migrated
     
class MigrateForVariable(Migrator):
    """Fix bug where variable value has new lines, and hence spills over to next line
       This will remove place newline back onto the same line
    """
    def __init__(self, list_of_input_lines):
        Migrator.__init__(self, list_of_input_lines)
 
    def _migration_hook(self,default_version_number):
        """
        The correct format is where variable in on the same line
            edit EMOS_TYPE '  family prod_wparam
  endfamily
'         
        """
        #count = 0;
        no_of_label_quotes = 0
        migrated = False
        start_line_append = ""
        for line in self.list_of_input_lines:
            #count =  count + 1
            #print(str(count) , ": " , line)
            #if count == 383119:
            #    print("debug mee")
            if len(start_line_append) > 0:
                no_of_label_quotes += line.count("'")
                start_line_append += line
                if no_of_label_quotes % 2 == 0:
                    # even number of quotes
                    self.list_of_output_lines.append(start_line_append) # preserve last '\n'  
                    start_line_append = ""
                    no_of_label_quotes = 0 
                else:
                    start_line_append = start_line_append.rstrip('\n')
                continue            
            else:
            
                # process line, avoid ECF_URL_CMD as that could by error have uneven number of "'' i.e
                #   edit ECF_URL_CMD '${BROWSER:=firefox} -new-tab %URL%'
                # hence we use 
                #    if no_of_label_quotes == 1: 
                # instead of
                #    if no_of_label_quotes % 2 == 0:
                tokens = line.split()
                if len(tokens) >= 3 and tokens[0] == "edit":
                    no_of_label_quotes = line.count("'")
                    if no_of_label_quotes == 1: 
                        start_line_append = line
                        start_line_append = start_line_append.rstrip('\n')
                        migrated = True
                        continue
                  
            self.list_of_output_lines.append(line)
           
        return migrated
     
     
class MigrateForHistory(Migrator):
    """Fix bug where history spans multiple lines.  
       See File: history_bug.def
    """
    def __init__(self, list_of_input_lines):
        Migrator.__init__(self, list_of_input_lines)
 
    def _migration_hook(self,default_version_number):
        """
        The history should all be on one line
            defs_state MIGRATE state>:queued flag:message state_change:1500811 modify_change:48
            edit ECF_LOG '/vol/emos_nc/output/ecflow/vali.21801.ecf.log'
            edit SMSNAME '0'
            history /s2s_devel/ecmf/back MSG:[15:53:06 21.10.2014] --replace=/s2s_devel/ecmf/back s2s_devel.def parent  :emos
            history /s2s_devel/ecmf/enfh MSG:[13:45:53 12.11.2014] -alter change label last_run 20140929
            20141002
            20141016
            20141023 /s2s_devel/ecmf/enfh/  :emos    MSG:[14:05:53 14.11.2014] --requeue force /s2s_devel/ecmf/enfh  :emos
            history /s2s_devel/ecmf/enfh/back MSG:[13:38:30 11.11.2014] --alter change event doIt set /s2s_devel/ecmf/enfh/back  :emos
            suite limits #  begun:1 state:complete flag:message suspended:1
            endsuite
        """
        migrated = False
        history_started = False
        for line in self.list_of_input_lines:
            # process line
            tokens = line.split()
            if len(tokens) >= 2 and tokens[0] == "suite":
                history_started = False

            if not history_started and len(tokens) >= 2 and tokens[0] == "history":
                history_started = True
 
            if history_started:
                # append to previous line
                if len(tokens) >= 1 and tokens[0] != "history":
                    # get last line of output lines, string the new line and append
                    self.list_of_output_lines[-1] = self.list_of_output_lines[-1].rstrip()
                    self.list_of_output_lines[-1] = self.list_of_output_lines[-1] + line
                    migrated = True
                    continue;
                  
            self.list_of_output_lines.append(line)
           
        return migrated  
    
    
def do_migrate(defs_file):
    
    migration_count = 0
    fileName, fileExtension = os.path.splitext(defs_file)
    output_file_name = fileName + ".mig";
    
    input_file_object = open(defs_file,'r')
    output_file_object = open(output_file_name,'w')
    try:
        list_of_input_lines = input_file_object.readlines()  # preserves new lines, we want this
        
        list_of_output_lines = []
        abort_migrator = MigrateForTaskAbort(list_of_input_lines)
        if abort_migrator.migrate(401):
            migration_count += 1
            # did migration, update input lines for next migration
            list_of_input_lines = abort_migrator.output_lines()
        list_of_output_lines = abort_migrator.output_lines()
         
         
        label_migrator = MigrateForLabel(list_of_input_lines)
        if label_migrator.migrate(319):
            migration_count += 1
            # did migration, update input lines for next migration
            list_of_input_lines = label_migrator.output_lines()
        list_of_output_lines = label_migrator.output_lines()


        variable_migrator = MigrateForVariable(list_of_input_lines)
        if variable_migrator.migrate(401):
            migration_count += 1
            # did migration, update input lines for next migration
            list_of_input_lines = variable_migrator.output_lines()
        list_of_output_lines = variable_migrator.output_lines()
        
        
        variable_migrator = MigrateForHistory(list_of_input_lines)
        if variable_migrator.migrate(406):
            migration_count += 1
            # did migration, update input lines for next migration
            list_of_input_lines = variable_migrator.output_lines()
        list_of_output_lines = variable_migrator.output_lines()
            
        output_file_object.writelines(list_of_output_lines)
    finally:
        input_file_object.close()
        output_file_object.close()
        
    return migration_count 

def usage():
    DESC = """\nThis python file is used to *FIX* migration bugs, when migrating from one release of ecflow to another.

 Migration may be required when the release number changes:
 <release-number>.<major>.<minor>
    6.1.7   ->  7.0.0
 as the checkpoint files *may* not be compatible. 
        
 This can be *checked* before hand. By attempting to load the check pt into the new server. 
 On the old server run:
    ecflow_client --check_pt  # this will write out a checkpt file  
 Now copy this file to a separate directory, and start the *new* server using the same port number
 If no errors are reported, you do not need to migrate.  
 If loading the checkpt file into the new server fails for any reason then
 then please follow these steps.

 Steps for Old server:
   o shutdown
        # ecflow_client --shutdown
   o suspend all suites
        # CL="ecflow_client --port=32222 --host=vsms1"
        # for s in $($CL --suites); do $CL --suspend=/$s; done
   o wait for active/submitted tasks to complete
   o halt the server
        # ecflow_client --halt=yes
   o Use --migrate to dump state and structure to a file
        # ecflow_client --migrate > all_suites.def
   o terminate server *or* leave server running but start new server on different machine
     to avoid port number clash.
   o remove checkpt and backup checkpt files, to prevent new server from loading them
     *Only* applicable if starting new server on same machine
   o Run this script 
     python ecflow_migrate.py -d all_suites.def
     This will fix any issues associated with reloading this defs into the new server.
     It write a file called 'all_suites.mig', which we will need to load into the new server. 
     See below.

 Steps for New server:
   o start server
   o load the migration file
       # ecflow_client --load all_suites.mig
   o set server running:
       # ecflow_client --restart
   o resume suspended suites
       # CL="ecflow_client --port=32222 --host=vsms1"
       # for s in $($CL --suites); do $CL --resume=/$s; done


 There could be bugs with *old* "ecflow_client --migrate > all_suites.def"
 This file, will fix any bugs associated --migrate, so that the subsequent
 load with  *new* client (ecflow_client --load all_suites.mig) will work

 Usage:
    python ecflow_migrate.py --defs_file <filename>
    python ecflow_migrate.py --d <filename>
    python ecflow_migrate.py --defs_file all_suites.def 
    
    The fixed definition is written as filename.mig, i.e. all_suites.mig
    """
    return DESC;
    
    
if __name__ == "__main__":
    
#     PARSER = argparse.ArgumentParser(description=usage(),  
#                                      formatter_class=argparse.RawDescriptionHelpFormatter)
#     PARSER.add_argument('defs_file',  help="The definition file to *fix* for migration")
#     ARGS = PARSER.parse_args()
#     ARGS.defs_file = os.path.expandvars(ARGS.defs_file) # expand references to any environment variables
#     print(ARGS)    
#     do_migrate( ARGS.defs_file )
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hd", ["help", "defs_file="])
    except getopt.GetoptError as err:
        # print("help information and exit:")
        print(err) # will print something like "option -a not recognised"
        print(usage())
        sys.exit(2)
        
    defs_file = None
    for o, a in opts:
        if o in ("-h", "--help"):
            print(usage())
            sys.exit()
        elif o in ("-d", "--defs_file"):
            defs_file = a
        else:
            assert False, "un-handled option"
    
    if defs_file is None:
        print("Please enter path to the defs file i.e ecflow_migrate --d ./defs_file.def")
        sys.exit(1)
        
    defs_file = os.path.expandvars(defs_file); # expand references to any environment variables
    do_migrate( defs_file )



# Copyright 2009-2019 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import sys

import ecflow

try:
    # Create the client
    ci = ecflow.Client("localhost", "4143")

    # Get the node tree suite definition as stored in the server
    # The definition is retrieved and stored on the variable 'ci'
    ci.sync_local()

    # access the definition retrieved from the server
    defs = ci.get_defs()

    if defs is None or len(defs) == 0:
        print("The server has no suites")
        sys.exit(1)

    # get the tasks, *alternatively* could use defs.get_all_nodes()
    # to include suites, families and tasks.
    task_vec = defs.get_all_tasks()

    # iterate over tasks and print path and state
    for task in task_vec:
        print(task.get_abs_node_path() + " " + str(task.get_state()))

except RuntimeError as e:
    print("Failed: " + str(e))



#!/usr/bin/env python

"""
python docstring: demonstrate python cli + ecFlow task wrapper + task loader
 
$manual
DESCRIPTION:
 
sweeper:
 
find out min/max among modeleps_nemo tasks
 
OPERATORS: please, set complete if problematic
 
ANALYST: example as pure python task - job
$end
 
$comment
  comments can be added ...
$end
 
"""
import time
import os.path
import getopt
import sys

PATH = "/usr/local/apps/ecflow/4.0.9/lib/python2.7/site-packages/ecflow"
sys.path.insert(0, PATH)
import ecflow as ec

MICRO = "$$"  # double dollar to please ecFlow micro character balance
TARGET_TASK = "modeleps_nemo"  # use option -t to overwrite
TARGET_METER = "step"  # would this need an option???
MC_STEPS_LIST = (
    "/mc/main/12/legA/fc",  # use option -p for single path
    "/mc/main/12/legB/fc",
    "/mc/main/18bc/legA/fc",
    "/mc/main/00/legA/fc",
    "/mc/main/00/legB/fc",
    "/mc/main/06bc/legA/fc",
    "/mofc/thu/01/legC/fc",
    "/mofc/thu/hind/14/back",
    "/mofc/mon/01/legC/fc",
    "/mofc/mon/hind/14/back",
)


def focus_below(node):
    """continue down below nodes which name is expected ;
    please update for external use...
    """
    return (
        node.name() in ("00", "12", "18bc", "06bc", "legA", "legB", "pf", "cf", "fc")
        or "main" == node.name()
    )


def process(item, path=None, low=999, high=-1, task=None):
    """track nodes to follow and update min/max"""
    if type(item) == str:
        defs = ec.Defs(item)
    elif isinstance(item, ec.Client):
        item.sync_local()
        defs = item.get_defs()
    elif isinstance(item, ec.Defs):
        defs = item
    else:
        defs = None

    if defs:
        for node in defs.suites:
            if node.name() != str(path.split("/")[1]):
                continue
            status = "%s" % node.get_state()
            if node.is_suspended() or status == "unknown":
                continue
            for item in node.nodes:
                low, high = process(item, path, low, high, task)

    elif isinstance(item, ec.Family):
        for node in item.nodes:
            if focus_below(node):
                return process(node, path, low, high, task)
            low, high = process(node, path, low, high, task)

    elif isinstance(item, ec.Task):
        node = item
        path2node = item.get_abs_node_path()
        status = "%s" % node.get_state()

        # ignore cal-val # ecmwf specific
        if "/cv/" in path2node or not task in item.name():
            return low, high
        if "/cv/" in path2node and (high > 240 or low < -1):
            high = 240
            return low, high

        stamp = ""
        evt = ""
        meter = ""
        if node.is_suspended():
            status = "suspended"
        try:
            for att in node.meters:
                meter = "%s" % att.value()
                if att.name() != TARGET_METER:
                    continue
                if att.value() > high:
                    high = att.value()
                elif att.value() < low:
                    low = att.value()

        except Exception as excpt:
            print("#! problem with line:", excpt)
        # name = "%20s %s %5s %5s" % (path2node, stamp, evt, meter)

    else:
        print(type(item))

    return low, high


def create_task():
    """add ecFlow task into a family..."""
    sys.path.append("/home/ma/emos/o/def")
    import ecf

    steps = 360
    return ecf.Task("sweeper").add(
        ecf.Variables(
            ECF_FILES=os.getenv("ECF_FDI", None),
            ECF_EXTN=".py",
            ECF_JOB_CMD="ssh -x $WSHOST$ $ECF_JOB$ > $ECF_JOBOUT$ 2>&1 &",
            ECF_MICRO=MICRO[0],
        ),
        ecf.Label("info", ""),
        ecf.Defcomplete(),
        ecf.Meter("min", -1, steps, steps),
        ecf.Meter("max", -1, steps, steps),
    )


def loader(test=1, host=os.getenv("ECF_TEST_HOST", None)):
    """replace sweeper task in test/oper mode"""
    sys.path.append("/home/ma/emos/o/def")
    import ecf

    if test:
        env = {
            "host": os.getenv("ECF_TEST_HOST", None),
            "port": os.getenv("ECF_TEST_PORT", None),
            "path": "/admine/steps",
        }
    else:
        env = {
            "host": os.getenv("ECF_OPER_HOST", None),
            "port": os.getenv("ECF_OPER_PORT", None),
            "path": "/admin/steps",
        }
    defs = ecf.Defs()

    path = env["path"]
    sname = str(path.split("/")[1])
    fname = str(path.split("/")[2])

    suite = ecf.Suite(sname).add(
        ecf.Extern(MC_STEPS_LIST, defs),
        ecf.Family(fname).add(
            create_task().add(
                ecf.Cron("04:30 23:59 03:00"),
                ecf.Trigger("==active or ".join(MC_STEPS_LIST) + "==active"),
                ecf.Variables(WSHOST=host, QUEUE="test"),
            )
        ),
    )

    defs.add_suite(suite)

    print("#MSG: replacing", path, env["host"], env["port"])
    ecf.Client(env["host"], env["port"]).replace(path, defs, 1, 1)


class Sweep(object):
    """find min/max meter value below a node"""

    def __init__(self, host=None, port=None, path=None, task=None):
        import signal

        if host is None:
            host = os.getenv("ECF_NODE", "$ECF_NODE$")
        if port is None:
            port = int(os.getenv("ECF_PORT", "$ECF_PORT$"))
        self.clt = ec.Client(host, port)
        self.cl2 = None
        if path is None:
            self.path = MC_STEPS_LIST
        else:
            self.path = path
        if task is None:
            self.task = TARGET_TASK
        else:
            self.task = task

        # shall make sense when processed into job by ecflow
        # name remains when not processed...
        ecfv = {
            "ECF_NODE": "$ECF_NODE$",
            "ECF_PASS": "$ECF_PASS$",
            "ECF_NAME": "$ECF_NAME$",
            "ECF_PORT": "$ECF_PORT$",
            "ECF_TRYNO": "$ECF_TRYNO$",
        }

        for key, val in ecfv.items():
            if key in val:
                ecfv[key] = os.getenv(key, None)
        if ecfv["ECF_NODE"] and ecfv["ECF_NAME"]:
            print("#MSG will communicated with server...")
            print("#kill: ssh %s kill -15 %d" % (ecfv["ECF_NODE"], os.getpid()))
            self.cl2 = ec.Client()
            self.cl2.set_host_port(ecfv["ECF_NODE"], ecfv["ECF_PORT"])
            self.cl2.set_child_pid(os.getpid())
            self.cl2.set_child_path(ecfv["ECF_NAME"])
            self.cl2.set_child_password(ecfv["ECF_PASS"])
            self.cl2.set_child_try_no(int(ecfv["ECF_TRYNO"]))
            self.cl2.child_init()
            self.cl2.set_child_timeout(20)

            for sig in (
                signal.SIGINT,
                signal.SIGHUP,
                signal.SIGQUIT,
                signal.SIGILL,
                signal.SIGTRAP,
                signal.SIGIOT,
                signal.SIGBUS,
                signal.SIGFPE,
                signal.SIGUSR1,
                signal.SIGUSR2,
                signal.SIGPIPE,
                signal.SIGTERM,
                signal.SIGXCPU,
                signal.SIGPWR,
            ):
                signal.signal(sig, self.signal_handler)

    def signal_handler(self, signum, frame):
        """catch signal"""
        print("Aborting: Signal handler called with signal ", signum)
        self.cl2.child_abort("Signal handler called with signal " + str(signum))

    def report(self, msg, meter=None):
        """communicate with ecFlow server"""
        if not self.cl2:
            return
        if meter:
            self.cl2.child_meter(msg, meter)
        elif msg == "stop":
            self.cl2.child_complete()
            sys.exit(0)
        else:
            self.cl2.child_label("info", msg)

    def update(self):
        """refresh status tree asking ecFlow server"""
        num = 20
        leg = "legA"
        for path in self.path:
            first = 1
            low = 0
            high = 0
            self.clt.ch_register(False, [str(path.split("/")[1])])
            if "/back" in path:
                path += "/%02d/%s" % (num, leg)

            while 1:
                clt = self.clt
                status = ""
                try:
                    if clt.news_local() or first:  # has the server changed
                        clt.sync_local()
                        first = 0
                        defs = clt.get_defs()
                        node = defs.find_abs_node(path)
                        if node:
                            status = "%s" % node.get_state()
                        low, high = process(node, self.path, 999, -1, self.task)
                        sleep = 30
                    else:
                        sleep = 90
                    msg = "... %ds %s %s %d %d" % (sleep, path, status, low, high)
                    print(msg)
                    if type(low) != type(high):
                        print(type(low), type(high))

                    if status == "complete":
                        if leg == "legB":
                            num -= 1
                            leg = "legA"
                        else:
                            leg = "legB"
                        break

                    elif (
                        low == high
                        and status in ("queued", "aborted", "complete")
                        or node is None
                    ):
                        if node:
                            msg = "# %s %s" % (node.get_abs_node_path(), status)
                            print(msg)
                            self.report(msg)
                        break
                    else:
                        self.report(msg)
                        self.report("min", low)
                        self.report("max", high)
                    sys.stdout.flush()
                    time.sleep(sleep)
                except RuntimeError as exp:
                    print(str(exp))


def usage():
    """help"""
    print(
        """client
  -o: operational node is to be replaced, default is test node,
      provide this option BEFORE -r
  -r: replace task node
  -p: path to look below, by default internal list MC_STEPS_LIST will be used
  -h: this help
 
ECF_NODE=localhost ECF_PORT=31415 ./client.py --path /mc/main/18bc/legA/fc -e
 
"""
    )


if __name__ == "__main__":
    try:
        OPTS, ARGS = getopt.getopt(
            sys.argv[1:], "hep:rot:", ["help", "ens", "path", "replace", "oper", "task"]
        )
    except getopt.GetoptError as err:
        print("# what?", usage())
        sys.exit(2)

    TEST = 1
    PATH = None
    TASK = None
    for o, a in OPTS:
        if o in (
            "-e",
            "--ens",
        ):
            Sweep(path=PATH).update()
            sys.exit(0)
        elif o in (
            "-r",
            "--replace",
        ):
            loader(TEST)
            sys.exit(0)
        elif o in (
            "-p",
            "--path",
        ):
            PATH = a
        elif o in (
            "-o",
            "--oper",
        ):
            TEST = 0
        elif o in (
            "-t",
            "--task",
        ):
            TASK = a
    CLT = Sweep(path=PATH, task=TASK)
    CLT.update()
    CLT.report("stop")
